;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2004 Mike Byrne
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : actr-interface.lisp
;;; Version     : 2.2
;;; 
;;; Description : Formerly "act-interface.lisp"
;;;             : This file contains functions which are ACT-R specific,
;;;             : particularly in how DM is handled.
;;; 
;;; Bugs        : [] No way to send SCALE info with default move-attention
;;;             :    command.
;;; 
;;; Todo        : 
;;; 
;;; ----- History -----
;;; 01.07.27 mdb
;;;             : Started 5.0 conversion work.
;;;             : [x] MOVE-ATTENTION can fail to generate a chunk {this might
;;;             :    go into the actr-interface.lisp}
;;; 01.08.07 mdb
;;;             : Added a chunk type for START-TRACKING.  Added SIZE slot on
;;;             : VISUAL-LOCATION chunk type.
;;; 01.09.17 mdb
;;;             : * Fixed glitch in :after methods for new-*-sound.
;;;             : * Returned MIN-RUN-TIME.
;;; 02.01.21 mdb
;;;             : Changed the proc-state part of state-spec.
;;; 
;;; 02.02.14 Dan
;;;             : Added a new pm-run to fix the "problem" with the variable
;;;             : thresholds in ACT-R.  Also had to modify the cognitive
;;;             : module accordingly...  Basically, matching-cycle sets the
;;;             : globals for rt and ut and action-cycle needs them and relies
;;;             : on the globals being set, but somewhere the "original"
;;;             : values need to be saved because matching-cycle needs them
;;;             : passed in as parameters, and the "original" value can be
;;;             : changed between calls to run (because people write some
;;;             : perverse models, and of course ACT-R isn't going to impose
;;;             : any modeling constraints on the user or even throw a warning
;;;             : for many badly formed productions - sorry a bit of a rant
;;;             : there). I'm sure there's a better solution, but this is the
;;;             : fastest solution (basically it's how Christian's run does
;;;             : it) which is what I need right now...
;;; 2002.04.10 mdb
;;;             : Added DISTANCE slot to visual location chunks.
;;; 2002.05.08 mdb [b6]
;;;             : UPDATE-TRACKING-MTH now updates the visual location buffer.
;;; 2002.05.23 mdb
;;;             : Modified sound chunks a little.
;;; 2002.06.24 mdb [b7]
;;;             : Another minor modification to sound events and added
;;;             : 'internal' and 'external' pm-constant chunks.
;;; 2003.01.21 mdb [2.1.1]
;;;             : Made PS-PTR a method.
;;; 2004.03.03 mdb [2.2]
;;;             : PROCESS-DISPLAY :after method killed, moved to VISICON-UPDATE.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;; ---------------------------------------------------------------------- ;;;;
;;;; Functions that define the relevant chunk types and such for ACT-R.
;;;; ---------------------------------------------------------------------- ;;;;

(defun def-pm-chunk-types ()
  "Defines chunk types used by the PM layer"
  (unless (get-type 'visual-location)
    
    ;; define the dmetype for state chunks
    (chunk-type module-state module modality processor 
                preparation execution last-command)
    
    ;; visual interface chunks
    (chunk-type visual-object screen-pos value status color height width)
    (chunk-type abstract-object value line-pos bin-pos)
    (chunk-type (abstract-letter (:include abstract-object)))
    (chunk-type (abstract-number (:include abstract-object)))
    (chunk-type (text (:include visual-object)))
    (chunk-type (empty-space (:include visual-object)))
    
    #| This is really never used and requires lists, so this was removed for 2.1
    (chunk-type (sub-object (:include visual-object)) objects)
    (chunk-type (point (:include sub-object)))
    (chunk-type (pair (:include sub-object)))
    (chunk-type (triplet (:include sub-object)))
    |#
    
    (chunk-type (line (:include visual-object)) other-pos end1-x end1-y end2-x end2-y)
    (chunk-type (oval (:include visual-object)))
    (chunk-type (cursor (:include visual-object)))
    
    (chunk-type (phrase! (:include visual-object)) objects words)
    (chunk-type visual-location screen-x screen-y distance attended kind color 
                value size nearest)
    (chunk-type vision-command)
    (chunk-type motor-command)
    (chunk-type audio-command)
    (chunk-type pm-constant)
    (chunk-type color)
    
    ;; audio chunks
    (chunk-type audio-event attended onset offset pitch kind location)
    (chunk-type sound kind content event)
    
    ;;; Adds the chunk types to hold the various actions as subtypes of motor-command
    ;;; Note that move-mouse is currently a chunk of type motor-command but the
    ;;; documented motor commands (in 1.0 manual) such as press-key etc are just plain chunks.
    (chunk-type (click-mouse (:include motor-command)))
    (chunk-type (hand-to-mouse (:include motor-command)))
    (chunk-type (hand-to-home (:include motor-command)))
    (chunk-type (move-cursor (:include motor-command)) object loc device)
    (chunk-type (peck (:include motor-command)) hand finger r theta)
    (chunk-type (peck-recoil (:include motor-command)) hand finger r theta)
    (chunk-type (point-hand-at-key (:include motor-command)) hand to-key)
    (chunk-type (press-key (:include motor-command)) key)
    (chunk-type (punch (:include motor-command)) hand finger)
    ;;; Also for speech command(s)
    (chunk-type speech-command)
    (chunk-type (speak (:include speech-command)) string)
    (chunk-type (subvocalize (:include speech-command)) string)
    ;;; Also for visual commands
    (chunk-type (start-tracking (:include vision-command)) object)
    ))


;;; add some state chunks to DM
(defun add-pm-state-chunks ()
  "Adds chunks for various RPM states."
  (unless (no-output (dm vision-state))
    (no-output
     (add-dm-fct
      '((free isa chunk)
        (busy isa chunk)
        (vision-state isa module-state module :VISION modality FREE processor FREE
                      preparation FREE execution FREE)
        (motor-state isa module-state module :MOTOR modality FREE processor FREE
                     preparation FREE execution FREE)
        (speech-state isa module-state module :SPEECH modality FREE processor FREE
                      preparation FREE execution FREE)
        (audio-state isa module-state module :AUDIO modality FREE processor FREE
                     preparation FREE execution FREE)
        
        (move-mouse isa motor-command)
        (click-mouse isa motor-command)
        (find-location isa vision-command)
        (move-attention isa vision-command)
        (punch isa motor-command)
        (press-key isa motor-command)
        (listen-for isa audio-command)
        (black isa color)
        (digit isa chunk)
        (speech isa chunk)
        (tone isa chunk)
        (oval isa chunk)
        (click-mouse isa chunk)
        (text isa chunk)
        (box isa chunk)
        (move-cursor isa chunk)
        (move-attention isa chunk)
        (punch isa chunk)
        (peck isa chunk)
        (peck-recoil isa chunk)
        (hand-ply isa chunk)
        (point-hand isa chunk)
        (press-key isa chunk)
        (point-hand-at-key isa chunk)
        (hand-to-mouse isa chunk)
        (hand-to-home isa chunk)
        (:motor isa chunk)
        (:vision isa chunk)
        (:audio isa chunk)
        (:speech isa chunk)
        (lowest isa pm-constant)
        (highest isa pm-constant)
        (greater-than-current isa pm-constant)
        (less-than-current isa pm-constant)
        (current isa pm-constant)
        (external isa pm-constant)
        (internal isa pm-constant)
        )
      :reset-ia nil))))



(def-pm-chunk-types)
(add-pm-state-chunks)


;;;; ---------------------------------------------------------------------- ;;;;
;;;; ACT-R methods for the Declarative Memory Interface
;;;; ---------------------------------------------------------------------- ;;;;

(defmethod ps-ptr ((self dmo))
  (get-safe-wme (id self)))


(defmethod make-dme (id (dmtype symbol) (attrs list) &key (obj nil) 
                          (where nil))
  (declare (ignore where))
  (no-output (add-dm-fct `((,id isa ,dmtype ,@attrs)) :reset-ia nil))
  (make-instance 'dmo :dmtype dmtype :id id :pm-obj obj))


(defmethod get-attribute ((self dmo) (attrname symbol))
  (no-output (chunk-slot-value-fct (ps-ptr self) attrname)))


(defmethod set-attributes ((self dmo) (attrs list))
  (no-output (mod-chunk-fct (ps-ptr self) attrs))
  self)


(defmethod psdme-to-dmo (psdme)
  (setf psdme (get-safe-wme psdme))
  (make-instance 'dmo :dmtype (wme-type psdme) :id (wme-name psdme)))


(defmethod dmo-to-psdme ((self dmo))
  (no-output (get-safe-wme (id self))))

;  (ps-ptr self))


(defmethod matching-dmos ((dmtype symbol) (attrs list))
  (let ((psdmes 
         (no-output (sdm-fct `(isa ,dmtype ,@attrs)))))
    (when psdmes
      (mapcar #'psdme-to-dmo psdmes))))

#|
;; mdb:  What's this doing here?  
(defmethod xy-to-dmo :around ((loc list) (attended-p symbol))
  (let ((the-dmo (call-next-method)))
    (no-output
     (add-reference (wme-references (dmo-to-psdme the-dmo)))
     (update-activation-spread))
    the-dmo))
|#


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Method overrides and specializers for ACT-R
;;;; ---------------------------------------------------------------------- ;;;;

;;; could this be an :after method?
(defmethod find-location :around ((vis-mod vision-module) &key (kind :IGNORE)
                                     (attended :IGNORE) (value :IGNORE)
                                     (color :IGNORE) (size :IGNORE)
                                     screen-x screen-y distance nearest)
  (declare (ignore kind attended value color size screen-x screen-y distance 
                   nearest))
  (aif (call-next-method)
    (setf *visual-location* it)
    (setf *visual-location* (get-wme 'FAILURE))))


(defmethod new-sound-event :around ((evt sound-event))
  (let ((evt (call-next-method)))
    (pm-timed-event (detect-at-time evt) #'stuff-sound-buffer)))


(defun stuff-sound-buffer  () 
  (unless *aural-location*
    (find-sound-fct :attended nil)))

;;; again, coul this be :after?
(defmethod find-sound :around ((aud-mod audio-module) 
                                 &key (attended :IGNORE) (kind :IGNORE) 
                                 onset pitch)
  (declare (ignore attended kind onset pitch))
  (aif (call-next-method)
    (setf *aural-location* it)
    (setf *aural-location* (get-wme 'FAILURE))))


(defmethod visicon-update :after ((vis-mod vision-module))
  (stuff-visloc-buffer vis-mod))


(defmethod min-run-time ((mstr-proc master-process))
  *default-action-time*)


(defmethod update-tracking-mth :after ((vm vision-module))
  (setf *visual-location* (dmo-to-psdme (current-marker vm))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Functions to interface with ACT-R's modified production syntax 
;;;; ---------------------------------------------------------------------- ;;;;

;;; vision

(defun current-visual-object ()
  "Return the chunk corresponding to the currently-attended object."
  (aif (currently-attended (vis-m *mp*))
    (dmo-to-psdme it)
    (get-wme 'failure)))                ; mdb 01.07.27


(defun find-location-fct (&key (kind :IGNORE) (attended :IGNORE)
                                  (value :IGNORE) (color :IGNORE)
                                  (size :IGNORE) (screen-x :IGNORE)
                                  (screen-y :IGNORE) nearest)
  "Calls the Vision Module's FIND-LOCATION method"
  (setf kind (check-gwn kind))
  (setf attended (check-gwn attended))
  (setf value (check-gwn value))
  (setf color (check-gwn color))
  (setf size (check-gwn size))
  (setf screen-x (check-gwn screen-x))
  (setf screen-y (check-gwn screen-y))
  (setf nearest (check-gwn nearest))
  (find-location (vis-m *mp*) :kind kind  :attended attended  :value value
                 :color color  :size size  :screen-x screen-x  
                 :screen-y screen-y :nearest nearest))


(defmethod stuff-visloc-buffer ((vis-mod vision-module))
  (unless *visual-location*
    (awhen (find-current-locs-with-spec vis-mod (default-spec vis-mod))
      (setf *visual-location*
            (construct-location vis-mod 
                                (random-item (objs-max-slotval it 'tstamp))
                                (default-spec vis-mod))))))



;;; audition


(defun find-sound-fct (&rest params)
  "Call's the Audio Manager's FIND-SOUND method."
  (apply #'find-sound (audio-m *mp*) params))




(defun current-sound-fct ()
  "Return the chunk corresponding to the currently-attended object."
  (awhen (currently-attended (audio-m *mp*))
    (dmo-to-psdme it)))


;;; module state

(defun module-state-fct (&key module (modality :ignore) 
                                 (preparation :ignore) (execution :ignore)
                                 (processor :ignore) 
                                 (last-command :ignore))
  "Function to check the state of a PM module, for the ACT-R LHS test."
  (when (match-spec-p (make-instance 'pm-state-spec 
                        :modality modality
                        :preparation preparation
                        :execution execution
                        :processor processor
                        :last-command last-command)
                      (key->mgr *mp* module))
    (dmo-to-psdme (state-dmo (key->mgr *mp* module)))))



(defclass pm-state-spec (pm-module spec)
  ()
  (:default-initargs
    :check-slots #(modality-state preparation-state execution-state 
                                  processor-state last-command)
    :modality :IGNORE
    :preparation :IGNORE
    :execution :IGNORE
    :processor :IGNORE
    :last-command :IGNORE))


#|
(defmethod match-spec-p :around ((ts pm-state-spec)
                                   (mod pm-module))
  (when (call-next-method)
    (let ((condition (proc-stub ts))
          (value (proc-s mod)))
      (cond ((eq condition :IGNORE) T)
            ((functionp condition)
             (funcall condition value))
            (t (equal condition value))))))


(defmethod match-spec-p :around ((ts pm-state-spec)
                                   (mod pm-module))
  (call-next-method))
|#



(defun pm-run (duration &key (full-time nil))
  "Runs ACT-R/PM for the specifed duration (in seconds).
   If full-time is t then it gurantees that time is advanded by duration seconds"
  (setf (rt (key->mgr *mp* :cognition)) *retrieval-threshold*)
  (setf (ut (key->mgr *mp* :cognition)) *utility-threshold*)
  (when full-time
    (pm-delayed-event duration #'(lambda () 
                                   (pm-output nil "* Running stopped because time limit reached."))))
  
  (multiple-value-prog1 
    (run-master-process *mp* duration)
    (setf  *retrieval-threshold* (rt (key->mgr *mp* :cognition)))
    (setf *utility-threshold* (ut (key->mgr *mp* :cognition)))))



#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#