;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Authors     : Christian Lebiere, Mike Byrne & Dan Bothell
;;; Address     : Carnegie Mellon University
;;;             : Psychology Department
;;;             : Pittsburgh,PA 15213-3890
;;;             : db30+@andrew.cmu.edu
;;; 
;;; Copyright   : (c)1998-2004 Christian Lebiere
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : actr5-pm-compiler-fix.lisp
;;; Version     : 5.1
;;;
;;; Description : Integration of the ACT-R/PM components with the 
;;;               production system.
;;;
;;; Bugs        : Buffer names don't matter on the RHS - only the
;;;             : chunk-type matters.
;;; 
;;; ----- History -----
;;; 04.04.13 Dan [5.1]
;;;             : Added the header, the copyright notice and the LGPL stuff.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
(defparameter *pm-calls* 
  '((visual-location
     find-location-fct
     (attended :attended)
     (screen-x :screen-x)
     (screen-y :screen-y)
     (value :value)
     (kind :kind)
     (color :color)
     (size :size)
     (nearest :nearest)
     (distance :distance)               ; mdb 2002.04.10
     (time))
    (audio-event
     find-sound-fct
     (attended :attended)
     (onset :onset)
     (offset :offset)
     (pitch :pitch)
     (kind :type)
     (objects)
     (time))
    (module-state
     module-state-fct
     (module :module)
     (preparation :preparation)
     (processor :processor)
     (execution :execution)
     (modality :modality)
     (last-command :last-command))
    (visual-object current-visual-object)
    (sound current-sound-fct))
  "Mapping between chunk types and fcts and if applicable slot names
   and keyword arguments for calls to the perceptual/motor layer.")
 
(defun sub-eval-vars (expr bindings &optional (fct-call t))
  "Substitute stack reference for production variables in evaled expressions.
   Return a function to be funcalled rather than a form to be evaled, which for all
   but the simplest ones causes the compiler to be called every time, unless
   *compile-eval-calls* has been turned off."
  (let ((new-expr (list (copy-tree expr))))
    (dolist (binding bindings (if fct-call
                                (if *compile-eval-calls*
                                  (compile nil `(lambda () ,(first new-expr)))
                                  (eval `(function (lambda () ,(first new-expr)))))
                                (first new-expr)))
      (nsubst `(get-wme-name (svref *instantiation* ,(variable-index binding)))
              (variable-name binding) new-expr :test #'eq))))
 
(defun compile-lhs (production lhs initializations bindings index)
  "Compiles the left-hand side of production.
   Add the instantiation-adding call."
  (let ((code nil)
        (form nil))
    (dolist (clause lhs)
      (let ((key (pop clause))
            (type nil)
            (tests nil)
            (binds nil)
            (slot nil)
            (value nil)
            (negation nil))
        (setf form
              (cond
               ((retrievalp key)
                (setf key (var>var key))
                (with-binding binding direct key bindings index
                  (set-variable-type production type key (pop clause))
                  (when type
                    (cond ((and (member (wme-type-name type) '(visual-location audio-event) :test #'eq)
                                    (member '(time now) clause :test #'equal)) ;; compatibility with the old hack
                           (setf clause (delete '(time now) clause :test #'equal))
                           (setf (variable-type binding) type)
                           (let* ((fct-args (rest (assoc (wme-type-name type) *pm-calls* :test #'eq)))
                                  (pm-call (list (first fct-args)))
                                  (argument-alist (rest fct-args)))
                             (dolist (slot-value clause)
                               (let* ((negation (when (equal (first slot-value) '-)
                                                  (pop slot-value)))
                                      (slot (first slot-value))
                                      (value (second slot-value))
                                      (mapping (assoc slot argument-alist)))
                                 (cond (mapping
                                        (when (and (or (not (variablep value))
                                                       (get-variable-binding value bindings))
                                                   (second mapping))
                                          (setf value (sub-eval-vars value bindings nil))
                                          (nconc pm-call (list (second mapping)
                                                               (if negation
                                                                 (list 'pm-not-equal value)
                                                                 value)))))
                                       (t
                                        (signal-warn "SLOT ~S OF CHUNK ~S TYPE ~S IN PRODUCTION ~S IS UNDEFINED."
                                                     slot key type production)))))
                             (compile-lhs-slots tests binds)
                             (list (if (and *enable-rational-analysis* code)
                                     ; the goal cannot be partial-matched
                                     'external-test-and-bind-pm
                                     'external-test-and-bind)
                                   (variable-index binding)
                                   pm-call
                                   type nil binds)))
                          ((or (member key '(=visual-state =aural-state =manual-state =vocal-state)) ;; reserved keywords for states
                               (eq (wme-type-name type) 'module-state))
                           (setf clause (delete '(time now) clause :test #'equal))
                           (setf (variable-type binding) type)
                           (let* ((fct-args (rest (assoc (wme-type-name type) *pm-calls* :test #'eq)))
                                  (pm-call (list (first fct-args)))
                                  (argument-alist (rest fct-args)))
                             (when (and (member key '(=visual-state =aural-state =manual-state =vocal-state))
                                        (not (assoc 'module clause)))
                               (nconc pm-call (list :module (case key
                                                              (=visual-state :vision)
                                                              (=aural-state :audio)
                                                              (=manual-state :motor)
                                                              (=vocal-state :speech)))))
                             (dolist (slot-value clause)
                               (let* ((negation (when (equal (first slot-value) '-)
                                                  (pop slot-value)))
                                      (slot (first slot-value))
                                      (value (second slot-value))
                                      (mapping (assoc slot argument-alist)))
                                 (cond (mapping
                                        (when (and (or (not (variablep value))
                                                       (get-variable-binding value bindings))
                                                   (second mapping))
                                          (setf value (sub-eval-vars value bindings nil))
                                          (nconc pm-call (list (second mapping)
                                                               (if negation
                                                                 (list 'pm-not-equal value)
                                                                 value)))))
                                       (t
                                        (signal-warn "SLOT ~S OF CHUNK ~S TYPE ~S IN PRODUCTION ~S IS UNDEFINED."
                                                     slot key type production)))))
                             
                             (compile-lhs-slots tests binds)
                             (list 'buffer-test-and-bind ;;  It is all a buffer match now
                                   (variable-index binding)
                                   pm-call
                                   type nil binds)))
                          ((or (member key '(=visual =aural)) ;; reserved keywords for visual and sound objects
                               (and (or (member 'visual-object (wme-type-supertypes type)
                                                :test #'eq :key #'wme-type-name)
                                        (member 'sound (wme-type-supertypes type)
                                                :test #'eq :key #'wme-type-name))
                                    (member '(time now) clause :test #'equal)))
                           (setf clause (delete '(time now) clause :test #'equal))
                           (setf (variable-type binding) type)
                           (compile-lhs-slots tests binds)
                           (list 'buffer-test-and-bind ;;  It is all a buffer match now
                                 (variable-index binding)
                                 (list (second (assoc (if (or (member 'visual-object (wme-type-supertypes type)
                                                                      :test #'eq :key #'wme-type-name)
                                                              (eq key '=visual))
                                                        'visual-object 'sound)
                                                      *pm-calls* :test #'eq)))
                                 type tests binds))
                          ((member key '(=manual =manual-location =vocal =vocal-location))
                           (signal-warn "RESERVED KEYWORD NOT YET IMPLEMENTED."))
                          ((assoc (var=var key) *buffers*)   ;;; lhs buffer matching
                           (let ((buffer (assoc (var=var key) *buffers*)))
                             (setf (variable-type binding) type)
                             (compile-lhs-slots tests binds)
                             (list 'buffer-test-and-bind
                                   (variable-index binding)
                                   (list (buffer-calls-equal-lhs (third buffer))
                                         (list (first buffer) (second buffer)))
                                   type tests binds)))
                          (t  ;; includes =visual-location and =aural-location because they are now buffers
                           (setf (variable-type binding) type)
                           (compile-lhs-slots tests binds)
                           (list (if (assoc (variable-index binding) initializations :test #'=)
                                   ;; Fixed buffers have their special test to avoid activation
                                   'direct-test-and-bind-buffer
                                   (if *enable-rational-analysis*
                                     (if direct 'direct-test-and-bind-pm
                                         'indirect-test-and-bind-pm)
                                     (if direct 'direct-test-and-bind
                                         'indirect-test-and-bind)))
                                 (variable-index binding) type tests binds))))))
               ((eq '!eval! key)
                (list 'eval-test (sub-eval-vars (first clause) bindings)))
               ((eq '!bind! key)
                (with-binding binding bound (first clause) bindings index
                  (when bound
                    (signal-warn "VARIABLE ~S ALREADY BOUND IN PRODUCTION ~A."
                                 (first clause) production))
                  (list 'bind-lhs (variable-index binding)
                        (sub-eval-vars (second clause) bindings))))
               ((eq '!find-location! key)
                (with-binding binding bound (first clause) bindings index
                  (when bound
                    (signal-warn "VARIABLE ~S ALREADY BOUND IN PRODUCTION ~A."
                                 (first clause) production))
                  (list 'bind-find-location (variable-index binding)
                        (sub-eval-vars (cons 'list (quote-arguments (rest clause)))
                                       bindings))))
               (t (signal-warn "UNKNOWN COMMAND ~S IN PRODUCTION ~A."
                               key production))))
        (setf code (nconc code form))))
    ;; FIX: split into goal and retrievals when ERA is enabled.
    (values (if *enable-rational-analysis*
              (let ((direct-length (first-retrieval-index code initializations)))
                (cons (nconc (subseq code 0 direct-length) (list 'add-instantiation-to-conflict-set))
                      (nconc (subseq code direct-length) (list 'not))))
              (nconc code (list 'add-instantiation-to-conflict-set)))
            bindings index)))
 
;;; added back for compatibility with the old time now hack
(defun external-test-and-bind (arguments)
  "Implements a direct match.  Tests for type and then slots, then bind
   variables.  Finally, call the function implementing the next clause.
   Arguments is a list of the wme's stack index, its type, a list of slots
   tests and bindings, and the next call if any.
   Returns t if a complete match is found, nil otherwise."
  (let* ((wme-index (pop arguments))
         ;; FIX: get the current visual or sound object by calling the PM layer
         ;; If the variable is already bound, test that it is identical
         (external-call (pop arguments))
         (wme (apply (first external-call) (rest external-call)))
         (binding (instantiation-variable *instantiation* wme-index))
         (wmetype (pop arguments))
         (slots-tested (pop arguments))
         (slots-bound (pop arguments)))
    (cond ((and binding (not (eq wme binding)))
           (signal-output *exact-matching-trace* "Current visual object ~S does not match existing binding ~S."
                          wme binding))
          (t
           (unless binding
             (setf (instantiation-variable *instantiation* wme-index) wme))
           (if (and (wmep wme) (subtype (wme-type wme) wmetype))
             (let ((slots (wme-slots wme)))
               ;; Set bindings before testing slots
               (and (set-bindings slots slots-bound)
                   ;; FIX: The retrieval threshold does not apply to PM calls
                    (let ((*retrieval-threshold* nil))
                      (signal-output *activation-trace* "Sources of activation are: ~S" *activation-sources*)
                      (test-slots wme slots slots-tested)) ;; wme-index))
                    (simulate-call arguments)))
             (signal-output *exact-matching-trace* "~S is not a CHUNK of type ~S."
                            wme wmetype))))))
 
(defun external-test-and-bind-pm (arguments)
  "Implements a direct match.  Tests for type and then slots, then bind
   variables.  Finally, call the function implementing the next clause.
   Arguments is a list of the wme's stack index, its type, a list of slots
   tests and bindings, and the next call if any.
   This is the rational analysis version, which handles partial matching too."
  (let* ((wme-index (pop arguments))
         ;; FIX: get the current visual or sound object by calling the PM layer
         ;; If the variable is already bound, test that it is identical
         (external-call (pop arguments))
         (wme (apply (first external-call) (rest external-call)))
         (binding (instantiation-variable *instantiation* wme-index))
         (wmetype (pop arguments))
         (slots-tested (pop arguments))
         (slots-bound (pop arguments)))
    (cond ((and binding (not (eq wme binding)))
           (signal-output *exact-matching-trace* "Current visual object ~S does not match existing binding ~S."
                          wme binding))
          ((and (wmep wme) (subtype (wme-type wme) wmetype))
           (unless binding
             (setf (instantiation-variable *instantiation* wme-index) wme))
           (let* ((slots (wme-slots wme))
                  ;; Set bindings before testing slots
                  (activation (and (set-bindings slots slots-bound)
                                   ;; FIX: The retrieval threshold does not apply to PM calls
                                   (let ((*retrieval-threshold* most-negative-short-float))
                                     (signal-output *activation-trace* "Sources of activation are: ~S" *activation-sources*)
                                     (if *partial-matching*
                                       (test-slots-pm wme slots slots-tested)
                                       (test-slots wme slots slots-tested))))))
             ;; FIX: Do not increment the latency
             ;; (add-latency wme (or activation *retrieval-threshold*))
             (when activation 
               (simulate-call arguments))))
          (t
           ;; FIX: Do not increment the latency
           ;; (add-latency wme *retrieval-threshold*)
           (signal-output *exact-matching-trace* "~S is not a CHUNK of type ~S."
                          wme wmetype)))))
 
(defparameter *rhs-pm-calls*
  '((visual-object
     (:vision move-attention)
     (screen-pos :location)
     (scale :scale))
    (visual-location
     (:vision find-location)
     (attended :attended)
     (screen-x :screen-x)
     (screen-y :screen-y)
     (value :value)
     (kind :kind)
     (color :color)
     (size :size)
     (nearest :nearest)
     (distance :distance)
     (time))
    (start-tracking                     ; 01.11.10 mdb
     (:vision start-tracking)           ; missing entry added [again]
     (object nil))
    (assign-finst                       ; 2003.02.06 mdb 
     (:vision assign-finst)             ; added new term here
     (object :object)
     (location :location))
    (click-mouse
     (:motor click-mouse))
    (hand-to-mouse
     (:motor hand-to-mouse))
    (hand-to-home
     (:motor hand-to-home))
    (move-cursor
     (:motor move-cursor)
     (object :object)
     (loc :loc)
     (device :device))
    (peck
     (:motor peck)
     (hand :hand)
     (finger :finger)
     (r :r)
     (theta :theta))
    (peck-recoil
     (:motor peck-recoil)             ; fixed 02.08.08 Dan
     (hand :hand)
     (finger :finger)
     (r :r)
     (theta :theta))
    (point-hand-at-key
     (:motor point-hand-at-key)
     (hand :hand)
     (to-key :to-key))
    (press-key
     (:motor press-key)
     (key nil))
    (punch
     (:motor punch)
     (hand :hand)
     (finger :finger))
    (speak
     (:speech speak)
     (string :text))                    ; 01.11.10 mdb fixed entry
    (subvocalize                        ; 2002.05.23 added entry
     (:speech subvocalize)
     (string :text))
    (sound
     (:audio attend-sound)
     (event :event))
    (audio-event
     (:audio find-sound)
     (attended :attended)
     (kind :kind)                       ; 2003.10.23 mdb fixed entry
     (location :location)               ; 2003.10.24 mdb added entry 
     (onset :onset)
     (pitch :pitch)
)
    ))
 
(defmethod not-equal ((criterion number))
  (lambda (val)
    (not (equal val criterion))))
 
(defmethod not-equal ((criterion list))
  (not-equal (check-fct criterion)))
 
(defun translate-comparison-tests (clause)
  (let ((new-clause nil))
    (dolist (pair clause)
      (if (test-modifier (first pair))
        (let ((modifier (first pair)))
          (cond ((or (eq modifier '>) (eq modifier '>=))
                 (push-last (list (second pair) (list 'greater-than (third pair))) new-clause))
                ((or (eq modifier '<) (eq modifier '<=))
                 (push-last (list (second pair) (list 'less-than (third pair))) new-clause))
                ((eq modifier '-)
                 (push-last (list (second pair) (list 'not-equal (third pair))) new-clause))
                (t (signal-warn "ILLEGAL TEST ~S APPLIED TO ~S ~S IN PRODUCTION ACTION."
                                modifier (second pair) (third pair)))))
        (push-last pair new-clause)))
    (do ((rest-clause new-clause (rest rest-clause)))
        ((null rest-clause))
      (let* ((first-pair (first rest-clause))
             (other-pair (assoc (first first-pair) (rest rest-clause) :test #'eq)))
        (when other-pair
          (let ((arguments nil))
            (when (listp (second first-pair))
              (setf arguments (append arguments (second first-pair))))
            (when (listp (second other-pair))
              (setf arguments (append arguments (second other-pair))))
            (when (and (member 'less-than arguments :test #'eq)
                       (member 'greater-than arguments :test #'eq))
              (rplacd first-pair (list (list 'within (min (second arguments) (fourth arguments))
                                             (max (second arguments) (fourth arguments)))))
              (setf rest-clause (delete other-pair rest-clause)))))))
    new-clause))
 
(defun compile-rhs (production rhs bindings index)
  "Compiles the right-hand side of production."
  (let ((code nil)
        (returns nil)
        (goal-stack (list (variable-name (first bindings)))))        
    (dolist (clause rhs)
      (let ((key (pop clause))
            (type nil)
            (assigns nil)
            (slot nil)
            (value nil)
            (negation nil))
        (setf code (nconc code
         (cond ((retrievalp key)
                (setf key (var>var key))
                (when (actionp key) (setf key (var=var key))) ;; strip out the +
                (cond ((clearp key)  ;; clearing buffers 
                       (setf key (var=var key))  ;; strip out the -
                       (cond ((member key '(visual manual aural vocal))  ;; clearing modules
                              (let ((command-call (list (case key 
                                                          (visual :vision)
                                                          (aural :audio)
                                                          (manual :motor)
                                                          (vocal :speech)) 'clear)))
                                (list (list 'action-command
                                            (sub-eval-vars (cons 'list (quote-arguments command-call))
                                                           bindings)))))
                             ((assoc key *buffer-keywords*)  ;; clearing ACT-R buffers
                              (list (list 'assign-buffer key nil)))
                             ((assoc key *buffers*)   ;;; rhs buffer clearing
                              (let ((buffer (assoc key *buffers*)))
                                (list (list (buffer-calls-minus-rhs (third buffer))
                                            (first buffer) (second buffer)))))
                             (t
                              (signal-warn "UNKNOWN BUFFER ~S CANNOT CLEAR." key))))
                      ((and (or (assoc key *buffer-keywords*)
                                ;; directly manipulate PM buffers
                                (member key '(visual manual vocal aural
                                              =visual =manual =vocal =aural))) 
                            (null (rest clause))
                            (null (rest (first clause)))
                            (listp (first (first clause))))
                       ;; assignment to buffer of result of function call
                       (list (list 'assign-buffer key
                                   (sub-eval-vars (first (first clause)) bindings))))
                      ((and (not (variablep key)) (assoc key *buffer-keywords*)
                            (null (rest clause)) (null (rest (first clause))))
                       ;; direct assignment to buffers, including goal focus and direct retrieval
                       (let ((binding (if (variablep (first (first clause)))
                                        (variable-index (get-variable-binding (first (first clause))
                                                                              bindings))
                                        (get-wme-or-constant (first (first clause))))))
                         (cond ((eq key 'goal)  ;; focus on goal
                                (setf goal-stack (butlast goal-stack))
                                (push-last key goal-stack)
                                (list (list 'focus-fct 
                                            binding
                                            (subgoal-returns key returns))))
                               ((eq key 'retrieval)   ;; direct retrieval
                                (list (list 'handle-failure
                                            (if *enable-rational-analysis*
                                              'direct-test-and-bind-pm
                                              'direct-test-and-bind)
                                            binding
                                            type nil nil
                                            'assign-retrieval
                                            binding)))
                               (t  ;; other buffer assignment
                                (list (list 'assign-buffer key binding))))))
                      (t
                       (with-binding binding modify key bindings index
                         (cond (modify
                                (setf type (variable-type binding))
                                (if (and type (not (integerp type)))
                                  (when (eq 'isa (caar clause))
                                    (set-variable-type production type key (pop clause))
                                    (unless (or (eq type (variable-type binding))
                                                (assoc key *buffer-keywords*))
                                      (signal-warn "TYPE OF VARIABLE ~S IS BEING REDEFINED IN PRODUCTION ~A."
                                                   key production)))
                                  (set-variable-type production type key (pop clause))))
                               (t
                                (set-variable-type production type key (pop clause))))
                         (when type
                           (cond ((or (eq 'retrieval key)
                                      (member (list '!retrieve! key) rhs :test #'equal))
                                  (let ((tests nil)
                                        (binds nil))
                                    (compile-lhs-slots tests binds)
                                    (list (list 'handle-failure
                                                (if *enable-rational-analysis*
                                                  (if modify
                                                    'direct-test-and-bind-pm
                                                    'indirect-test-and-bind-pm)
                                                  (if modify
                                                    'direct-test-and-bind
                                                    'indirect-test-and-bind))
                                                nil
                                                type tests binds
                                                'assign-retrieval
                                                nil))))
                                 ;;; Partial implementation of RHS buffers.
                                 ;;; Full implementation will require further standardization of RPM conventions
                                 ((member key '(visual manual visual-location vocal aural aural-location))
                                  (let* ((fct-args (rest (assoc (wme-type-name type) *rhs-pm-calls* :test #'eq)))
                                         (command-call (copy-seq (first fct-args)))
                                         (argument-alist (rest fct-args)))
                                    (dolist (slot-value (translate-comparison-tests clause))
                                      (let* ((slot (first slot-value))
                                             (value (second slot-value))
                                             (mapping (assoc slot argument-alist)))
                                        (if mapping
                                          (setf command-call
                                                (nconc command-call (if (second mapping) (list (second mapping) value)
                                                                        (list value))))
                                          (signal-warn "SLOT ~S OF CHUNK ~S TYPE ~S IN PRODUCTION ~S IS UNDEFINED."
                                                       slot key type production))))
                                    (list (list 'action-command
                                                (sub-eval-vars (cons 'list (quote-arguments command-call))
                                                               bindings)))))
                                 ((member key '(visual-state manual-location manual-state vocal-location vocal-state aural-state))
                                  (signal-warn "RESERVED KEYWORD ~A NOT YET IMPLEMENTED." key))
                                 ((assoc key *buffers*)  ;;; rhs buffer action
                                  (let ((buffer (assoc key *buffers*))
                                        (arguments (sub-eval-vars (cons 'list (quote-arguments (cons 'isa (cons type (leaves clause)))))
                                                                  bindings)))
                                    (list (list (buffer-calls-plus-rhs (third buffer))
                                                (first buffer) (second buffer)
                                                arguments))))
                                  ((assoc (var=var key) *buffers*)   ;;; rhs buffer modification
                                   (let ((buffer (assoc (var=var key) *buffers*))
                                         (arguments (sub-eval-vars (cons 'list (quote-arguments (leaves clause)))
                                                                  bindings)))
                                     (list (list (buffer-calls-equal-rhs (third buffer))
                                                 (first buffer) (second buffer) arguments))))
                                 (t
                                  (compile-rhs-slots assigns returns binding)
                                  (cons
                                   (if modify
                                     (list 'modify-old-wme (variable-index binding)
                                           type assigns)
                                     (list 'create-new-wme (variable-index binding)
                                           (if (variablep key) (var=var key) key) type assigns))
                                   (when (and (or (eq '=newgoal key) (eq 'newgoal key) (eq 'goal key))
                                              (not (member (list '!push! key) rhs :test #'equal))
                                              (not (member (list '!focus-on! key) rhs :test #'equal)))
                                     (setf goal-stack (butlast goal-stack))
                                     (push-last key goal-stack)
                                     (list (list 'focus-fct 
                                                 (variable-index binding)
                                                 (subgoal-returns key returns))))))))))))
               ((eq '!retrieve! key) ;; do nothing since retrieval already done
                nil)
               ((eq '!push! key)
                (let* ((subgoal (first clause))
                       (binding (get-safe-variable-binding subgoal bindings production)))
                  (when binding
                    (push-last subgoal goal-stack)
                    (list (list 'push-fct
                                (variable-index binding)
                                (subgoal-returns subgoal returns))))))
               ((eq '!pop! key)
                ;;; FIX: pop from the end of the stack
                (setf goal-stack (butlast goal-stack))
                (list (list 'pop-fct)))
               ((eq '!focus-on! key)
                (let* ((subgoal (first clause))
                       (binding (get-safe-variable-binding subgoal bindings production)))
                  (when binding
                    ;;; FIX: do not assume that there is something on the stack
                    (setf goal-stack (butlast goal-stack))
                    (push-last subgoal goal-stack)
                    (list (list 'focus-fct 
                                (variable-index binding)
                                (subgoal-returns subgoal returns))))))
               ((eq '!output! key)
                (list (list 'output (compile-output clause bindings production))))
               ((eq '!eval! key)
                (list (list 'eval-side (sub-eval-vars (first clause) bindings))))
               ((eq '!bind! key)
                (with-binding binding bound (first clause) bindings index
                  (when bound
                    (signal-warn "VARIABLE ~S ALREADY BOUND IN PRODUCTION ~A."
                                 (first clause) production))
                  (list (list 'bind-rhs (variable-index binding)
                              (sub-eval-vars (second clause) bindings)))))
               ((eq '!move-attention! key)
                (with-binding binding bound (first clause) bindings index
                  (when bound
                    (signal-warn "VARIABLE ~S ALREADY BOUND IN PRODUCTION ~A."
                                 (first clause) production))
                  (list (list 'bind-move-attention (variable-index binding)
                              (sub-eval-vars (cons 'list (quote-arguments (rest clause)))
                                             bindings)))))
               ((eq '!press-key! key)
                (list (list 'visual-action 'press-key
                            (sub-eval-vars (cons 'list (quote-arguments clause))
                                           bindings))))
               ((eq '!move-mouse! key)
                (list (list 'visual-action 'move-mouse
                            (sub-eval-vars (cons 'list (quote-arguments clause))
                                           bindings))))
               ((eq '!click-mouse! key)
                (list (list 'visual-action 'click-mouse
                            (sub-eval-vars (cons 'list (quote-arguments clause))
                                           bindings))))
               ((eq '!send-command! key)
                (list (list 'action-command
                            (sub-eval-vars (cons 'list (quote-arguments clause))
                                           bindings))))
               ((eq '!delete! key)
                (let ((binding (get-safe-variable-binding (first clause) bindings production)))
                  (when binding
                    (list (list 'delete-wme-variable (variable-index binding))))))
               ((or (eq '!copy! key) (eq '!copywme! key))
                (with-binding binding bound (first clause) bindings index
                  (when bound
                    (signal-warn "VARIABLE ~S ALREADY BOUND IN PRODUCTION ~A."
                                 (first clause) production))
                  (list (list 'copy-chunk-variable (variable-index binding)
                              (sub-eval-vars (cons 'list (rest clause))
                                             bindings)))))
               ((eq '!stop! key)
                (list (list 'stop)))
               ((eq '!restart! key)
                (list (list 'restart-top-goal)))
               (t (signal-warn "UNKNOWN COMMAND ~S IN PRODUCTION ~A." key production)))))))
    ; move retrieval clauses at the end to allow activation spreading
    (dolist (clause code)
      (when (eq (first clause) 'handle-failure)
        (setf code (nconc (delete clause code :test #'eq) (list clause)))))
    ; remove the unknown commands
    ; (setf code (delete nil code :test #'eq))
    ; translate wme variables into relative goal stack indices for value returns
    (dolist (return returns)
      (let ((from-position
             (position (variable-name (variable-index return))
                       goal-stack :test #'eq)))
        (if from-position
          (dolist (destination (variable-returns return))
            (let ((to-position (position (first destination) goal-stack :test #'eq)))
              (if (and (numberp to-position) (> from-position to-position))
                (rplaca destination (- from-position to-position 1))
                (signal-warn "VARIABLE ~A CANNOT BE RETURNED BECAUSE CHUNK ~A IS NOT ON THE STACK BELOW SUBGOAL ~A IN PRODUCTION ~A."
                       (variable-name return) (first destination)
                       (variable-name (variable-index return)) production))))
          (signal-warn "VARIABLE ~A CANNOT BE RETURNED BECAUSE CHUNK ~A IS NOT ON THE STACK IN PRODUCTION ~A."
                       (variable-name return)
                       (variable-name (variable-index return)) production))))
    (values code bindings index))) 

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#