;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2003 Mike Byrne
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : audio-module.lisp
;;; Version     : 2.1.3
;;; 
;;; Description : Source for RPM's Audition Module
;;; 
;;; Bugs        : 
;;; 
;;; Todo        : 
;;; 
;;; ----- History -----
;;; 
;;; 01.07.28 mdb
;;;             : Start of 2.1/ACT5 work
;;;             : [x] NEW-*-SOUND functions changed to methods.
;;;             : [x] Remvoved LISTEN-FOR {this is handled by buffer stuffing}
;;;             :     and cleaned up FIND-SOUND.
;;; 01.09.17 mdb
;;;             : Made some changes to support buffer stuffing more cleanly
;;;             : when time is noisy.
;;; 2002.04.16 mdb [b6]
;;;             : Addded a bunch of :INITFORM NIL to classes to prevent unbound
;;;             : slots.
;;; 2002.05.23 mdb
;;;             : Added word sound stuff.  Renamed chunks to be "Audio-Event"
;;;             : not "Sevt".
;;; 2002.06.24 mdb [b7]
;;;             : Added LOCATION slot to sound events.
;;; 2002.06.26 mdb
;;;             : Fixed teensy-weensy glitch with sound event initializer and
;;;             : event->dmo now includes location.
;;; 2003.10.24 mdb [2.1.3]
;;;             : * Changed name of BUILD-SEVT-DMO to AUDIO-ENCODING-COMPLETE.
;;;             : * Duration of word sounds now gotten from speech module. 
;;; 2003.11.21 mdb
;;;             : Fixed bug in EVENT->DMO.  Duh.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;



(defclass audio-module (attn-module)
  ((audicon :accessor audicon :initarg :audicon :initform nil)
   (digit-detect-delay :accessor digit-dtct-dly :initarg :digit-dtct-dly
                       :initform 0.300)
   (digit-recode-delay :accessor digit-rec-dly :initarg :digit-rec-dly
                       :initform 0.500)
   (digit-duration :accessor digit-duration :initarg :digit-duration 
                   :initform 0.600)
   (tone-detect-delay :accessor tone-dtct-dly :initarg :tone-dtct-dly
                      :initform 0.050)
   (tone-recode-delay :accessor tone-rec-dly :initarg :tone-rec-dly 
                      :initform 0.285)
   (sound-decay-time :accessor decay-time :initarg :decay-time
                     :initform 3.000)
   )
  (:default-initargs
    :version-string "2.1.3"
    :name :AUDIO))


(defmethod reset-module :after ((aud-mod audio-module))
  (setf (audicon aud-mod) nil))


(defmethod initialize-instance :after ((aud-mod audio-module) &key)
  (setf (state-dmo aud-mod)
        (make-dme 'audio-state 'module-state
                  '(module :audio modality free processor free preparation free
                    execution free)
                  :where :external)))


(defmethod silent-events ((aud-mod audio-module))
  (awhen (next-detectable-sound aud-mod (pm-time))
    (pm-timed-event it #'get-internal-run-time)))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Sound events.
;;;; ---------------------------------------------------------------------- ;;;;

(defclass sound-event ()
  ((onset :accessor onset :initarg :onset 
          :initform (when *mp* (mp-time *mp*)))
   (offset :accessor offset :initarg :offset :initform nil)
   (string :accessor snd-string :initarg :string :initform nil)
   (duration :accessor duration :initarg :duration :initform 0)
   (content :accessor content :initarg :content :initform nil)
   (content-delay :accessor delay :initarg :delay :initform nil)
   (kind :accessor kind :initarg :kind :initform 'SPEECH)
   (attended-p :accessor attended-p :initform nil :initarg :attended-p)
   (location :accessor location :initarg :location :initform 'EXTERNAL)
   (sname :accessor sname :initform (new-name-fct "SOUND") :initarg :sname)
   (ename :accessor ename :initform (new-name-fct "AUDIO-EVENT"))
   (recode :accessor recode :initarg :recode :initform nil)
   (pitch :accessor pitch :initform 'middle :initarg :pitch)
   (snd-dmo :accessor snd-dmo :initform nil)
   (evt-dmo :accessor evt-dmo :initform nil)
   ))


(defmethod initialize-instance :after ((self sound-event) &key)
  (unless (offset self)
    (when (and (numberp (onset self)) (numberp (duration self)))
      (setf (offset self) (+ (onset self) (duration self))))))

(defgeneric detect-at-time (evt)
  (:documentation "Returns the time at which an event becomes detectable."))

(defmethod detect-at-time ((evt sound-event))
  (ms-round (+ (onset evt) (delay evt))))


(defgeneric detectable-p (evt)
  (:documentation  "Returns T if the given sound event is detectable."))

(defmethod detectable-p ((evt sound-event))
  (>= (pm-time) (detect-at-time evt)))


(defgeneric finished-p (evt)
  (:documentation  "Returns T if the given sound-event is finished."))

(defmethod finished-p ((evt sound-event))
  (>= (mp-time *mp*) (offset evt)))


(defgeneric detectable-time (evt)
  (:documentation  "Returns the time at which the given sound event will become detectable."))

(defmethod detectable-time ((evt sound-event))
  (+ (onset evt) (delay evt)))


(defclass digit-sound-evt (sound-event)
  ()
  (:default-initargs
    :kind 'DIGIT
    :duration (when *mp* (rand-time (digit-duration (audio-m *mp*))))
    :delay (when *mp* (rand-time (digit-dtct-dly (audio-m *mp*))))
    :recode (when *mp* (digit-rec-dly (audio-m *mp*)))
    :sname (new-name-fct "DIGIT")))


(defmethod initialize-instance :after ((self digit-sound-evt) &key)
  (setf (content self) (snd-string self)))


;;; TONE-SOUND-EVENT      [Class]
;;; Date        : 97.04.03
;;; Description : Class for tone events.
;;;             : The CONTENT slot should be the tone frequency.

(defclass tone-sound-evt (sound-event)
  ()
  (:default-initargs
    :string ""
    :kind 'TONE
    :content 1000
    :delay (when *mp* (rand-time (tone-dtct-dly (audio-m *mp*))))
    :recode (when *mp* (tone-rec-dly (audio-m *mp*)))
    :sname (new-name-fct "TONE")))


(defmethod initialize-instance :after ((self tone-sound-evt) &key)
  (cond ((> (content self) 1500) (setf (pitch self) 'high))
        ((< (content self) 900) (setf (pitch self) 'low))))


(defclass word-sound-evt (sound-event)
  ()
  (:default-initargs
    :kind 'WORD
    :delay (when *mp* (rand-time (digit-dtct-dly (audio-m *mp*))))
    :sname (new-name-fct "WORD")
    :duration 0
    :recode nil
    ))

(defmethod initialize-instance :after ((self word-sound-evt) &key)
  (when (or (null (duration self)) (zerop (duration self)))
    (setf (duration self)
          (get-articulation-time (speech-m *mp*) (snd-string self))))
  (unless (recode self)
    (setf (recode self)
          ;; change the value below to make "hearing" faster
          (ms-round (max
                     (/ (duration self) 2)
                     (- (duration self) 0.150)))))
  (setf (content self) (snd-string self))
  (setf (offset self) (+ (onset self) (duration self)))
  )

(defclass sound-event-spec (sound-event spec)
  ()
  (:default-initargs
    :check-slots #(onset kind attended-p pitch location)
    :onset :IGNORE
    :kind :IGNORE
    :attended-p :IGNORE
    :pitch :IGNORE
    :offset :IGNORE
    :location :IGNORE
    ))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Toplevel commands.
;;;; ---------------------------------------------------------------------- ;;;;

(defgeneric new-sound-event (evt)
  (:documentation "Handles the bookkeeping when a new sound event is created."))

(defmethod new-sound-event ((evt sound-event))
  (push evt (audicon (audio-m *mp*)))
  evt)

(defgeneric new-digit-sound (digit &optional onset)
  (:documentation "Creates and adds a digit sound <digit>, a string, starting optionally at <onset>."))


(defmethod new-digit-sound ((digit number) 
                               &optional (onset (mp-time *mp*)))
  (new-sound-event (make-instance 'digit-sound-evt :onset (ms-round onset) 
                                  :string digit)))


(defgeneric new-tone-sound (freq duration &optional onset)
  (:documentation "Creates and adds a tone sound of <freq>, starting optionally at <onset>."))

(defmethod new-tone-sound ((freq number) (duration number)
                              &optional (onset (mp-time *mp*)))
  (new-sound-event (make-instance 'tone-sound-evt :onset (ms-round onset) 
                                  :duration duration :content freq)))

(defgeneric new-other-sound (content duration delay recode 
                                         &optional instr onset)
  (:documentation "Creates and adds a sound <content>, lasting <duration>, with content delay <delay>, with recode time <recode>, starting optionally at <onset>."))


(defmethod new-other-sound (content 
                               (duration number) (delay number)
                               (recode number) &optional (instr "")
                               (onset (mp-time *mp*)))
  (new-sound-event (make-instance 'sound-event :onset (ms-round onset) 
                                  :duration duration :content content 
                                  :delay delay :recode recode :string instr)))



(defgeneric new-word-sound (word &optional onset location)
  (:documentation "Creates and adds a word with optional onset time."))


(defmethod new-word-sound ((word string) 
                              &optional (onset (mp-time *mp*))
                              (location 'external))
  (new-sound-event (make-instance 'word-sound-evt :onset (ms-round onset) 
                                  :string word :location location)))



;;; FIND-SOUND      [Method]
;;; Date        : 97.08.18, delta 99.08.30
;;; Description : Parallels the Vision Module's FIND-LOCATION, this one finds
;;;             : audio events (not sounds) and returns a PS-specific DME.

(defgeneric find-sound (aud-mod &key attended kind onset pitch)
  (:documentation  "Given a set of specifications, return a sound event which matches."))

(defmethod find-sound ((aud-mod audio-module) &key (attended :IGNORE) 
                         (kind :IGNORE) onset pitch (location :ignore))
  (let ((event-ls nil)
        (found-evt nil)
        (spec (make-instance 'sound-event-spec
                 :attended-p attended  :kind kind :location location
                 :onset (if (or (null onset) (symbolp onset)) 
                          :IGNORE onset)
                 :pitch (if (or (null pitch) (symbolp pitch)) 
                          :IGNORE pitch))))
    ;; find features matching the spec
    (setf event-ls (objs-match-spec (detectable-audicon aud-mod) spec))
    ;; some filtering
    (case onset
      (lowest (setf event-ls (objs-min-slotval event-ls 'onset)))
      (highest (setf event-ls (objs-max-slotval event-ls 'onset))))
    (case pitch
      (lowest (setf event-ls (objs-min-slotval event-ls 'pitch)))
      (highest (setf event-ls (objs-min-slotval event-ls 'pitch))))
    (when event-ls
      (setf found-evt (random-item event-ls))
      (when found-evt
        (unless (evt-dmo found-evt)
          (event->dmo found-evt))
        (dmo-to-psdme (evt-dmo found-evt))))))


;;; ATTEND-SOUND      [Method]
;;; Date        : 97.08.18
;;; Description : Parallels the Vision Module's MOVE-ATTENTION, this one
;;;             : attends an audio event, ultimately building a chunk based
;;;             : on the content of the sound.

(defgeneric attend-sound (aud-mod &key event)
  (:documentation  "Shift auditory attention to the given sound event."))

(defmethod attend-sound ((aud-mod audio-module) &key event)
  (unless (check-jam aud-mod)
    (let ((event (find (wme-name (get-safe-wme event)) (audicon aud-mod)
                        :test #'(lambda (x y) (eq x (ename y))))))
      (setf (attended-p event) t)
      (setf (current-marker aud-mod) event)
      (change-state aud-mod :exec 'busy)
      (queue-command
       :time (recode event) :where :AUDIO :command 'audio-encoding-complete
       :randomize t :params event))))


#|
;;; LISTEN-FOR      [Method]
;;; Date        : 97.08.18, delta 99.08.30
;;; Description : Combination of FIND and ATTEND.  Does a FIND, and it if
;;;             : finds anything, immediately attends it.

(defgeneric listen-for (aud-mod &key onset kind attended pitch)
  (:documentation  "Checks the audicon for appropriate sounds.  If one is found, attend to it."))

(defmethod listen-for ((aud-mod audio-module) &key onset (kind :ignore) 
                         (attended :ignore) pitch)
  (multiple-value-bind (psdme found-evt)
                       (find-sound aud-mod :attended attended :kind kind
                                   :onset onset :pitch pitch)
    (declare (ignore psdme))
    (when found-evt
      (attend-event aud-mod found-evt))))
|#

;;;; ---------------------------------------------------------------------- ;;;;
;;;; support for toplevel commands
;;;; ---------------------------------------------------------------------- ;;;;


;;; ATTEND-EVENT      [Method]
;;; Date        : 97.04.11
;;; Description : When a sound is found by LISTEN-FOR, this may get called on
;;;             : the event.  This method handles state-setting and queueing
;;;             : of the appropriate actions.  Two situations are possible
;;;             : with the sound event:  the content is not yet available 
;;;             : [that is, the content-delay for the event has not passed]
;;;             : or it is.  If not, set preparation to busy until content
;;;             : becomes available.  Then, after recode time, actually add
;;;             : the item to declarative memory.

(defgeneric attend-event (aud-mod sevt)
  (:documentation  "When LISTEN-FOR picks up an event, this handles it."))

(defmethod attend-event ((aud-mod audio-module) (sevt sound-event))
  (let ((curr-time (mp-time *mp*))
        (detect-time (+ (onset sevt) (delay sevt))))
    (setf (current-marker aud-mod) sevt)
    (cond ((< curr-time detect-time)  		; sound not yet 'bufferized'
           (change-state aud-mod :prep 'busy)
           (queue-command
            :time (- detect-time curr-time) :where :AUDIO 
            :command 'change-state :params '(:exec busy :prep free))
           (queue-command
            :time (- (+ detect-time (rand-time (recode sevt))) curr-time)
            :where :AUDIO :command 'audio-encoding-complete
            :params sevt))
          (t
           (change-state aud-mod :exec 'busy)
           (queue-command
            :time (recode sevt) :where :AUDIO :command 'audio-encoding-complete
            :randomize t :params sevt)))))



(defgeneric audio-encoding-complete (aud-mod sevt)
  (:documentation  "Actually add a sound to declarative memory."))

(defmethod audio-encoding-complete ((aud-mod audio-module) 
                                         (sevt sound-event))
  (change-state aud-mod :exec 'free)
  (setf (attended-p sevt) t)
  (unless (snd-dmo sevt)
    (let ((the-dmo
           (make-dme (sname sevt) 'sound 
                     `(kind ,(kind sevt) content ,(content sevt)
                            event ,(ename sevt))
                     :where :external)))
      (setf (snd-dmo sevt) the-dmo)))
  (set-attended aud-mod (snd-dmo sevt)))



;;;; ---------------------------------------------------------------------- ;;;;
;;;; Audio utilities
;;;; ---------------------------------------------------------------------- ;;;;


(defgeneric purge-old-sounds (aud-mod)
  (:documentation  "Removes sounds that have decayed from the audicon"))

(defmethod purge-old-sounds ((aud-mod audio-module))
  (setf (audicon aud-mod)
        (remove-if #'(lambda (e) (< (+ (offset e) (decay-time aud-mod))
                                     (mp-time *mp*)))
                   (audicon aud-mod))))


(defun earliest-onset (evt-lis)
  "Returns the sound event in the list with earliest onset."
  (let ((best (onset (first evt-lis)))
        (outlis (list (first evt-lis))))
    (dolist (evt (rest evt-lis) outlis)
      (cond ((= (onset evt) best) (push evt outlis))
            ((< (onset evt) best) (setf best (onset evt))
             (setf outlis (list evt)))))))


(defun latest-onset (evt-lis)
  "Returns the sound event in the list with latest onset."
  (let ((best (onset (first evt-lis)))
        (outlis (list (first evt-lis))))
    (dolist (evt (rest evt-lis) outlis)
      (cond ((= (onset evt) best) (push evt outlis))
            ((> (onset evt) best) (setf best (onset evt))
             (setf outlis (list evt)))))))


(defgeneric current-audicon (aud-mod)
  (:documentation  "Returns the audicon, assuming all events that currently exist are in there."))

(defmethod current-audicon ((aud-mod audio-module))
  (purge-old-sounds aud-mod)
  (remove-if #'(lambda (x) (> (onset x) (mp-time *mp*))) (audicon aud-mod)))


(defgeneric detectable-audicon (aud-mod)
  (:documentation  "Returns the audicon, but only those events that are currently detectable."))

(defmethod detectable-audicon ((aud-mod audio-module))
  (purge-old-sounds aud-mod)
  (remove-if #'(lambda (x) (not (detectable-p x))) (audicon aud-mod)))



(defgeneric event->dmo (evt)
  (:documentation  "Translate a sound event to the corresponding DMO."))

(defmethod event->dmo ((evt sound-event))
  (let ((dmo
         (make-dme (ename evt) 'audio-event 
                   `(attended ,(attended-p evt) onset ,(onset evt)
                              location ,(location evt) kind ,(kind evt))
                   :where :external)))
    (when (finished-p evt)
      (set-attributes dmo `(offset ,(offset evt))))
    (setf (evt-dmo evt) dmo)))


(defgeneric next-detectable-sound (aud-mod current-time)
  (:documentation  "Return the time when the next sound in the audicon is detectable."))

(defmethod next-detectable-sound ((aud-mod audio-module) current-time)
  (let ((onsets (mapcar #'detectable-time (audicon aud-mod))))
    (setf onsets (sort onsets #'<))
    (dolist (event-time onsets nil)
      (when (> event-time current-time)
        (return-from next-detectable-sound event-time)))))


#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#