;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Authors     : Mike Byrne & Dan Bothell
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2003 Mike Byrne & Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LICENSE.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : device-interface.lisp
;;; Version     : 2.1.3
;;; 
;;; Description : File for managing the device interface.
;;; 
;;; Bugs        : None known.
;;; 
;;; Todo        : Nothing pending.
;;; 
;;; ----- History -----
;;; 01.09.21 mdb [b2]
;;;             : Changed PROCESS-DISPLAY and UPDATE-CURSOR-FEAT to generate
;;;             : fewer gentemps.  Probably faster, too.
;;; 01.11.11 mdb [b3]
;;;             : Added TRACE-MOUSE-P and MOUSE-TRACE slots.
;;; 
;;; 02.01.15 Dan [b4]
;;;             : Changed the feature test for ACL in the setting
;;;             : of the pixels-per-inch veriables so that it works
;;;             : in other than ACL 5.0.1.  
;;;             : I don't think those values even matter however because
;;;             : when pm-reset gets called the device interface 
;;;             : pixels-per-inch gets set to the value of the 
;;;             : corresponding parameter which defaults to 72, and
;;;             : not the current "real" values as defined by the
;;;             : system's resolution.
;;; 02.03.08 mdb [b5]
;;;             : PROCESS-DISPLAY now checks to see if a production is 
;;;             : executing, and if it is, schedules the proc for *after*
;;;             : the production is done.
;;; 2002.04.16 mdb [b6]
;;;             : Fixed glitch with the b5 change.
;;; 2002.05.22 mdb
;;;             : Made mouse takeover switchable.
;;; 2002.06.05 mdb
;;;             : Incorportated Dan's priority change fix.
;;; 02.06.09 mdb [b7]
;;;             : Put in Dan's RUN-EVENTS functions, which calls the hook.
;;; 02.06.21 Dan
;;;             : Changed the #+:mcl to better work with openmcl.
;;; 2002.07.01 mdb
;;;             : PROCESS-DISPLAY now correctly propagates CLEAR parameter
;;;             : when deferred.
;;; 2002.08.16 Dan [b8]
;;;             : Fixed an apparent bug in the defclass of device-interface -
;;;             : the accessor for the microphone-delay slot was :microphone-delay
;;;             : which caused a problem in LispWorks.  I changed it to
;;;             : microphone-delay.
;;; 2002.11.25 Dan [2.1.1]
;;;             : Added the :ACTR-ENV-ALONE switch to the ACL pixels-per-inch
;;;             : computations.
;;; 2003.10.23 mdb [2.1.3]
;;;             : 
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; unless it's MCL define the pixels-per-inch
#+(or (not :mcl) :openmcl)
(progn
  (defvar *pixels-per-inch-x* 72)
  (defvar *pixels-per-inch-y* 72))

;;; if it's ACL with the IDE (Windows) then find the true units-per-inch
#+(and :ALLEGRO-IDE (not :ACTR-ENV-ALONE))
(multiple-value-bind (x y) 
                     (stream-units-per-inch  (screen *system*))
  (setf *pixels-per-inch-x* x)
  (setf *pixels-per-inch-y* y))

;;;; ---------------------------------------------------------------------- ;;;;
;;; roll in everything that was in "environment-interface"

(defvar *actr-enabled-p* t)

(defun pm-get-time ()
  "If ACT is in control, return simulated time, else return real time (in ms)."
  (if *actr-enabled-p*
    (round (* 1000 (mp-time *mp*)))
    (get-internal-real-time)))



(defun pm-timed-event (time func &rest args)
  "To schedule an external event at a specific time (in seconds), call this"
  (setf (sched-q *mp*)
        (queue-insert (make-instance 'input-queue-entry
                        :time-tag time
                        :destination :EXTERNAL
                        :params (cons func args))
                      (sched-q *mp*))))


(defun pm-delayed-event (delay func &rest args)
  "To schedule an external event at some delay (in seconds) after the current time, call this"
  (setf (sched-q *mp*)
        (queue-insert (make-instance 'input-queue-entry
                        :time-tag (+ (mp-time *mp*) delay)
                        :destination :EXTERNAL
                        :params (cons func args))
                      (sched-q *mp*))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Base DEVICE-INTERFACE class and some quicky methods.

(defclass device-interface ()
  ((pixels-per-inch :accessor ppi :initarg :ppi
                    :initform (/ (+ *pixels-per-inch-x* 
                                    *pixels-per-inch-y*) 2.0))
   (viewing-distance :accessor viewing-distance :initform 15.0)
   (device :accessor device :initform nil)
   (key-closure-time :accessor key-closure-time :initform 0.010)
   (microphone-delay :accessor microphone-delay :initform 0.100)
   (keyboard :accessor keyboard :initform (make-instance 'virtual-keyboard))
   (speech-hook :accessor speech-hook :initform nil)
   (device-speech-p :accessor device-speech-p :initform nil)
   (with-cursor-p :accessor with-cursor-p :initform nil)
   (input-q :accessor input-q :initform nil)
   (device-hook :accessor device-hook :initform nil)
   (mouse-fitts-coeff :accessor mouse-fitts-coeff :initform 0.1)
   (show-focus-p :accessor show-focus-p :initarg :show-focus-p :initform nil)
   (trace-mouse-p :accessor trace-mouse-p :initarg :trace-mouse-p :initform nil)
   (mouse-trace :accessor mouse-trace :initarg :mouse-trace :initform nil)
   (needs-mouse-p :accessor needs-mouse-p :initarg :needs-mouse-p :initform t)
   (true-cursor-loc :accessor true-cursor-loc :initarg :true-cursor-loc
                    :initform #(0 0))
   (version-string :accessor version-string :initarg :version-string 
                   :initform "2.1.3")
   ))


(defmethod my-name ((mod device-interface))
  :DEVICE)

(defgeneric angle->pixels-mth (devin angle)
  (:documentation  "Determine the number of pixels subtending a visual angle."))

(defmethod angle->pixels-mth ((devin device-interface) (angle number))
  (round (* (* (viewing-distance devin) (tan (deg->rad angle))) 
            (ppi devin))))


(defgeneric pixels->angle-mth (devin pixels)
  (:documentation  "Determine the amount of visual angle subtended by <pixels>."))

(defmethod pixels->angle-mth ((devin device-interface) (pixels number))
  (rad->deg (atan (/ (/ pixels (ppi devin)) (viewing-distance devin)))))


(defgeneric find-viewing-dist-mth (devin angle pixels)
  (:documentation  "Given the number of pixels an angle subtends, what's the viewing distance?"))

(defmethod find-viewing-dist-mth ((devin device-interface) angle pixels)
  (floor 
   (/ pixels (* (tan (deg->rad angle)) (ppi devin)))))





;;;; ---------------------------------------------------------------------- ;;;;
;;;; Interacting with the Master Process


(defmethod reset-device ((devin device-interface))
  (setf (input-q devin) nil)
  (setf (mouse-trace devin) nil))

(defmethod reset-module ((devin device-interface))
  (reset-device devin))


(defgeneric update-device (devin time)
  (:documentation  "Update the device at <time>."))

;;; the default method has to do the following:
;;; [1] Update the attentional focus
;;; [2] Call the device hook
;;; [3] Make sure the cursor synch is maintained
;;; [4] Call the device update method

(defmethod update-device ((devin device-interface) time)
  (when (and (show-focus-p devin)
             (current-marker (vis-m *mp*)))
    (device-update-attended-loc (device devin) 
                                (dmo-to-xy (current-marker (vis-m *mp*)))))
  (awhen (device-hook devin)
    (funcall it time))
  (synch-mouse devin)
  (device-update (device devin) time))


(defgeneric device-update-attended-loc (device xyloc)
  (:documentation  "Tell the device to update the attended location, which is passed in."))

(defmethod device-update-attended-loc (device xyloc)
  (declare (ignore device xyloc))
  nil)


(defgeneric device-update (device time)
  (:documentation  "Update the device at <time>.  A method should be defined for this if the device is dynamic."))

(defmethod device-update (device time)
  (declare (ignore device time))
  nil)


(defgeneric run-events (devin the-time)
  (:documentation  "Runs the current events in the input queue."))

(defmethod run-events ((devin device-interface) the-time)
  (while (and (input-q devin)
              (>= (ms-round the-time)
                  (ms-round (time-tag (first (input-q devin))))))
    (let ((event (pop (input-q devin))))
      (when (functionp (step-hook *mp*))
        (funcall (step-hook *mp*) event))
      (when (trace-modules *mp*)
        (pm-output the-time "Device running command ~S"
                   (first (params event)) the-time))
      (apply (first (params event))
             (append (list devin) (rest (params event)))))))


(defmethod new-message ((devin device-interface) (entry input-queue-entry))
  (setf (input-q devin) (queue-insert entry (input-q devin))))


(defmethod pm-install-module ((mp master-process) (module device-interface)
                                 &optional (warn t))
  (awhen (assoc :DEVICE (module-lst mp))
    (awhen (and (neq module (rest it)) warn)
      (pm-warning "Overwrite current device interface?  [y/n] ")
      (let ((char (read-char)))
        (when (not (or (eq char #\y) (eq char #\Y)))
          (return-from pm-install-module nil))))
    (setf (module-lst mp) (delete :DEVICE (module-lst mp) :key #'first)))
  (push (cons :DEVICE module) (module-lst mp)))


;;; this just does nothing because the device gets updated before the
;;; schedule queue in DO-UPDATE.  We just need the stub because this call
;;; gets mapped on all the modules during the update, and we don't want the
;;; device updated twice.

(defmethod update-module ((devin device-interface))
  nil)


;;; bascially, same deal with this method
(defmethod silent-events ((devin device-interface))
  nil)


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Proc screen stuff

;;; PROCESS-DISPLAY      [Method]
;;; Description : Replacement for old PROC-SCREEN functionality.  A few things
;;;             : have to be done to process the display.  First, make sure
;;;             : a display is attached.  Then, set up the previous icon.
;;;             : This is NIL if we're clearing the display.  Then, ask the
;;;             : device for a new icon.  Then check the new icon against the
;;;             : old one so we don't lose state information.  Update the
;;;             : attended location, make sure everyone is current on the
;;;             : mouse location.  Return the length of the icon, since ya
;;;             : gotta return something.

(defgeneric process-display (devin vis-mod &optional clear)
  (:documentation  "Rebuild the Vision Module's icon based on the current display."))

(defmethod process-display ((devin device-interface) 
                               (vis-mod vision-module) &optional (clear nil))
  (let ((tmpicon nil)
        (time (pm-time)))
    ;; make sure a device is actually installed
    (unless (device devin)
      (pm-warning "Cannot process display--no device is installed.")
      (return-from process-display nil))
    ;; reschedule if a production is currently executing
    (when (eq (ps-state (cog-m *mp*)) :PENDING-EXECUTION)
      (schedule-post-module *mp* :COGNITION :command 'pm-proc-display 
                            :where :EXTERNAL :priority -1
                            :params `(:clear ,clear))
      (return-from process-display nil))
    ;; build the temp visicon
    (setf tmpicon (flatten (build-features-for (device devin) vis-mod)))
    (when (with-cursor-p devin)
      (awhen (cursor-to-feature (device devin))
        (push  it tmpicon)))
    ;; if clear, then make the visicon the temp icon, if not we have to
    ;; check the new vs. the old icon
    (if clear
      (setf (visicon vis-mod) (mapcar #'(lambda (f)
                                          (enter-into-visicon f time))
                                      tmpicon))
      (let ((oldicon (visicon vis-mod)))
        (setf (visicon vis-mod) nil)
        (dolist (feat tmpicon)
          (aif (member feat oldicon :test #'feat=)
            (progn
              (push (checknew (first it) vis-mod) (visicon vis-mod))
              (setf oldicon (delete (first it) oldicon)))
            (push (enter-into-visicon feat time) (visicon vis-mod))))
        (setf (visicon vis-mod) (nreverse (visicon vis-mod)))))
    ;; other bookkeeping 
    (synch-mouse devin)
    (visicon-update vis-mod)
    ))


(defgeneric synch-mouse (devin)
  (:documentation  "Make sure everyone agress on the current cursor position."))

(defmethod synch-mouse ((devin device-interface))
  (when (and (device devin)
             (needs-mouse-p devin)
             (not (vpt= (true-cursor-loc devin)
                        (get-mouse-coordinates (device devin)))))
    (device-move-cursor-to (device devin) (true-cursor-loc devin)))
  (update-cursor-feat devin (vis-m *mp*)))



(defgeneric update-cursor-feat (devin vis-mod)
  (:documentation  "Updates the feature in the icon with the current cursor position."))

(defmethod update-cursor-feat ((devin device-interface) 
                                  (vis-mod vision-module))
  (when (and (with-cursor-p devin) (device devin))
    (let ((new-pos (get-mouse-coordinates (device devin)))
          (cur-crsr (first (member 'CURSOR (visicon vis-mod) :key #'kind))))
      (if (null cur-crsr)
        (awhen (cursor-to-feature (device devin))
          (length (push (enter-into-visicon it (pm-time)) (visicon vis-mod))))
        (when (not (vpt= new-pos (xy-loc cur-crsr)))
          (setf (attended-p cur-crsr) nil)
          (setf (screen-x cur-crsr) (px new-pos))
          (setf (screen-y cur-crsr) (py new-pos)))))))




(defgeneric key->cmd (devin key)
  (:documentation  "Given a key, return the appropriate Motor Module command to type it."))

(defmethod key->cmd ((devin device-interface) key)
  (key-to-command (keyboard devin) key))


;;; Generic methods which will need to be overridden for specific devices.
;;; These are defined for MCL windows.

(defgeneric get-mouse-coordinates (device)
  (:documentation  "Return the mouse coordinates in #(x y) form."))

(defmethod get-mouse-coordinates (device)
  (error "No method defined for GET-MOUSE-COORDINATES on object ~S." device))

(defgeneric build-features-for (obj vis-mod)
  (:documentation  "Return a list of icon-features for an object."))

(defmethod build-features-for (obj vis-mod)
  (declare (ignore vis-mod))
  (error "No method defined for BUILD-FEATURES-FOR on object ~S." obj))

(defgeneric cursor-to-feature (device)
  (:documentation  "Reaturn an icon feature reprsenting the current cursor."))

(defmethod cursor-to-feature (device)
  (error "No method definded for CURSOR-TO-FEATURE on object ~S." device))  


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Output functions:  mouse movements, keypress stuff.


(defgeneric output-key (devin keyloc)
  (:documentation  "Request that the device register a key output for the key at a given location."))

(defmethod output-key ((devin device-interface) (keyloc vector))
  (let ((the-key (loc->key (keyboard devin) keyloc)))
    (if (eq the-key 'mouse)
      (device-handle-click (device devin))
      (device-handle-keypress (device devin) the-key))))


(defgeneric output-speech (devin text)
  (:documentation  "Requests that the device output the provided text as speech."))

(defmethod output-speech ((devin device-interface) (text string))
  (when (device-speech-p devin)
    (device-speak-string (device devin) text))
  (when (speech-hook devin)
    (funcall (speech-hook devin) (mp-time *mp*) text)))


(defgeneric move-cursor-absolute (devin xyloc)
  (:documentation  "Request that the device move the cursor to the absolute location <xyloc>."))

(defmethod move-cursor-absolute ((devin device-interface) xyloc)
  (when (trace-mouse-p devin)
    (push (cons (pm-time) xyloc) (mouse-trace devin)))
  (device-move-cursor-to (device devin) xyloc)
  (setf (true-cursor-loc devin) xyloc)
  (synch-mouse devin))


(defgeneric move-cursor-polar (devin rtheta)
  (:documentation  "Request that the device move the cursor by <r> in direction <theta>."))

(defmethod move-cursor-polar ((devin device-interface) rtheta)
  (let ((newloc (polar-move-xy (get-mouse-coordinates (device devin)) rtheta)))
    (device-move-cursor-to (device devin) newloc)
    (setf (true-cursor-loc devin) newloc))
  (synch-mouse devin))


(defgeneric output-click (devin)
  (:documentation  "Output a mouse click to the device."))

(defmethod output-click ((devin device-interface))
  (device-handle-click (device devin)))


;;; Generic methods which will need to be overridden for specific devices.
;;; These are defined for MCL windows.

(defgeneric device-handle-keypress (device key)
  (:documentation  "Handle the press of the given key."))

(defmethod device-handle-keypress (device key)
  (declare (ignore key))
  (error "No method defined for DEVICE-HANDLE-KEYPRESS for object ~S." device))

(defgeneric device-move-cursor-to (device xyloc)
  (:documentation  "Move the cursor to a specified location."))

(defmethod device-move-cursor-to (device xyloc)
  (declare (ignore xyloc))
  (error "No method defined for DEVICE-MOVE-CURSOR-TO for object ~S." device))

(defgeneric device-handle-click (device)
  (:documentation  "Handle a click request."))

(defmethod device-handle-click (device)
  (error "No method defined for DEVICE-HANDLE-CLICK for object ~S." device))

(defgeneric device-speak-string (device string)
  (:documentation  "Handle a speech request."))

(defmethod device-speak-string (device string)
  (declare (ignore string))
  (error "No method defined for DEVICE-SPEAK-STRING for object ~S.") device)


;;;; ---------------------------------------------------------------------- ;;;;
;;;; The virtual keyboard.  Override one or more of these methods to 
;;;; represent different keyboards.

(defclass virtual-keyboard ()
   ((key->cmd-ht :accessor key->cmd-ht 
                  :initform (make-hash-table :test #'equal))
    (key->loc-ht :accessor key->loc-ht
                  :initform (make-hash-table :test #'equal))
    (loc->key-array :accessor loc->key-arr 
                     :initform (make-array '(23 7) :initial-element nil))))


(defmethod initialize-instance :after ((vk virtual-keyboard) &key)
  (populate-key-to-command-ht (key->cmd-ht vk))
  (populate-key-to-loc-ht (key->loc-ht vk))
  (populate-loc-to-key-array (loc->key-arr vk)))


;;; LOC-TO-KEY      [Method]
;;; Description : Given a location, return the corresponding character.
;;;             : Accessed via the 'virtual keyboard' array.

(defgeneric loc->key (vk loc)
  (:documentation  "Given an location, return the corresponding key"))

(defmethod loc->key ((vk virtual-keyboard) (loc vector))
  (if (vpt= loc #(28 2))
    'mouse
    (aref (loc->key-arr vk) (px loc) (py loc))))


(defgeneric key-to-command (vk key)
  (:documentation  "Given a key, return the appropriate command."))

(defmethod key-to-command ((vk virtual-keyboard) key)
  (gethash key (key->cmd-ht vk)))


(defmethod key-to-loc ((vk virtual-keyboard) key)
  (gethash key (key->loc-ht vk)))


(defgeneric populate-key-to-command-ht (ht)
  (:documentation  "Populates the hash table that maps keys to motor commands"))

(defmethod populate-key-to-command-ht ((ht hash-table))
  (setf (gethash 'space ht) '(punch :hand left :finger thumb))
  (setf (gethash 'backquote ht) 
        '(peck-recoil :hand left :finger pinkie :r 2.24 :theta -2.03))
  (setf (gethash 'tab ht) 
        '(peck-recoil :hand left :finger pinkie :r 1.41 :theta -2.36))
  (setf (gethash '1 ht) 
        '(peck-recoil :hand left :finger pinkie :r 2 :theta -1.57))
  (setf (gethash 'Q ht) 
        '(peck-recoil :hand left :finger pinkie :r 1 :theta -1.57))
  (setf (gethash 'A ht) '(punch :hand left :finger pinkie))
  (setf (gethash 'Z ht) 
        '(peck-recoil :hand left :finger pinkie :r 1 :theta 1.57))
  (setf (gethash '2 ht) 
        '(peck-recoil :hand left :finger ring :r 2 :theta -1.57))
  (setf (gethash 'W ht) 
        '(peck-recoil :hand left :finger ring :r 1 :theta -1.57))
  (setf (gethash 'S ht) 
        '(punch :hand left :finger ring))
  (setf (gethash 'X ht) 
        '(peck-recoil :hand left :finger ring :r 1 :theta 1.57))
  (setf (gethash '3 ht) 
        '(peck-recoil :hand left :finger middle :r 2 :theta -1.57))
  (setf (gethash 'E ht) 
        '(peck-recoil :hand left :finger middle :r 1 :theta -1.57))
  (setf (gethash 'D ht) '(punch :hand left :finger middle))
  (setf (gethash 'C ht) 
        '(peck-recoil :hand left :finger middle :r 1 :theta 1.57))
  (setf (gethash '4 ht) 
        '(peck-recoil :hand left :finger index :r 2 :theta -1.57))
  (setf (gethash 'R ht) 
        '(peck-recoil :hand left :finger index :r 1 :theta -1.57))
  (setf (gethash 'F ht) '(punch :hand left :finger index))
  (setf (gethash 'V ht) 
        '(peck-recoil :hand left :finger index :r 1 :theta 1.57))
  (setf (gethash '5 ht) 
        '(peck-recoil :hand left :finger index :r 2.24 :theta -1.11))
  (setf (gethash 'T ht) 
        '(peck-recoil :hand left :finger index :r 1.41 :theta -0.79))
  (setf (gethash 'G ht) 
        '(peck-recoil :hand left :finger index :r 1 :theta 0))
  (setf (gethash 'B ht) 
        '(peck-recoil :hand left :finger index :r 1.41 :theta 0.79))
  (setf (gethash '6 ht) 
        '(peck-recoil :hand right :finger index :r 2.24 :theta -2.03))
  (setf (gethash 'Y ht) 
        '(peck-recoil :hand right :finger index :r 1.41 :theta -2.36))
  (setf (gethash 'H ht) 
        '(peck-recoil :hand right :finger index :r 1 :theta 3.14))
  (setf (gethash 'N ht) 
        '(peck-recoil :hand right :finger index :r 1.41 :theta 2.36))
  (setf (gethash '7 ht) 
        '(peck-recoil :hand right :finger index :r 2 :theta -1.57))
  (setf (gethash 'U ht) 
        '(peck-recoil :hand right :finger index :r 1 :theta -1.57))
  (setf (gethash 'J ht) 
        '(punch :hand right :finger index))
  (setf (gethash 'M ht) 
        '(peck-recoil :hand right :finger index :r 1 :theta 1.57))
  (setf (gethash '8 ht) 
        '(peck-recoil :hand right :finger middle :r 2 :theta -1.57))
  (setf (gethash 'I ht) 
        '(peck-recoil :hand right :finger middle :r 1 :theta -1.57))
  (setf (gethash 'K ht) '(punch :hand right :finger middle))
  (setf (gethash 'comma ht) 
        '(peck-recoil :hand right :finger middle :r 1 :theta 1.57))
  (setf (gethash '9 ht) 
        '(peck-recoil :hand right :finger ring :r 2 :theta -1.57))
  (setf (gethash 'O ht) 
        '(peck-recoil :hand right :finger ring :r 1 :theta -1.57))
  (setf (gethash 'L ht) 
        '(punch :hand right :finger ring))
  (setf (gethash 'period ht) 
        '(peck-recoil :hand right :finger index :r 1 :theta 1.57))
  (setf (gethash '0 ht) 
        '(peck-recoil :hand right :finger pinkie :r 2 :theta -1.57))
  (setf (gethash 'P ht) 
        '(peck-recoil :hand right :finger pinkie :r 1 :theta -1.57))
  (setf (gethash 'semicolon ht) '(punch :hand right :finger pinkie))
  (setf (gethash 'slash ht) 
        '(peck-recoil :hand right :finger pinkie :r 1 :theta 1.57))
  (setf (gethash 'hyphen ht) 
        '(peck-recoil :hand right :finger pinkie :r 2.24 :theta 1.11))
  (setf (gethash '[ ht) 
        '(peck-recoil :hand right :finger pinkie :r 1.41 :theta -0.78))
  (setf (gethash 'quote ht) 
        '(peck-recoil :hand right :finger pinkie :r 1 :theta 0))
  (setf (gethash 'return ht) 
        '(peck-recoil :hand right :finger pinkie :r 2 :theta 0))
  ht
  )

(defgeneric populate-key-to-loc-ht (ht)
  (:documentation  "Build a hash table mapping keys to locations."))

(defmethod populate-key-to-loc-ht ((ht hash-table))
  ;; function key row
  (setf (gethash 'escape ht) #(0 0))
  (setf (gethash 'F1 ht) #(2 0))
  (setf (gethash 'F2 ht) #(3 0))
  (setf (gethash 'F3 ht) #(4 0))
  (setf (gethash 'F4 ht) #(5 0))
  (setf (gethash 'F5 ht) #(7 0))
  (setf (gethash 'F6 ht) #(8 0))
  (setf (gethash 'F7 ht) #(9 0))
  (setf (gethash 'F8 ht) #(10 0))
  (setf (gethash 'F9 ht) #(12 0))
  (setf (gethash 'F10 ht) #(13 0))
  (setf (gethash 'F11 ht) #(14 0))
  (setf (gethash 'F12 ht) #(15 0))
  (setf (gethash 'F13 ht) #(17 0))
  (setf (gethash 'F14 ht) #(18 0))
  (setf (gethash 'F15 ht) #(19 0))
  ;; numeric key row
  (setf (gethash 'backquote ht) #(0 2))
  (setf (gethash 1 ht) #(1 2))
  (setf (gethash 2 ht) #(2 2))
  (setf (gethash 3 ht) #(3 2))
  (setf (gethash 4 ht) #(4 2))
  (setf (gethash 5 ht) #(5 2))
  (setf (gethash 6 ht) #(6 2))
  (setf (gethash 7 ht) #(7 2))
  (setf (gethash 8 ht) #(8 2))
  (setf (gethash 9 ht) #(9 2))
  (setf (gethash 0 ht) #(10 2))
  (setf (gethash '- ht) #(11 2))
  (setf (gethash '= ht) #(12 2))
  (setf (gethash 'delete ht) #(13 2))
  (setf (gethash 'help ht) #(15 2))
  (setf (gethash 'home ht) #(16 2))
  (setf (gethash 'pageup ht) #(17 2))
  (setf (gethash 'clear ht) #(19 2))
  (setf (gethash '= ht) #(20 2))
  (setf (gethash '/ ht) #(21 2))
  (setf (gethash '* ht) #(22 2))
  ;; QWERTY row
  (setf (gethash 'tab ht) #(0 3))
  (setf (gethash 'Q ht) #(1 3))
  (setf (gethash 'W ht) #(2 3))
  (setf (gethash 'E ht) #(3 3))
  (setf (gethash 'R ht) #(4 3))
  (setf (gethash 'T ht) #(5 3))
  (setf (gethash 'Y ht) #(6 3))
  (setf (gethash 'U ht) #(7 3))
  (setf (gethash 'I ht) #(8 3))
  (setf (gethash 'O ht) #(9 3))
  (setf (gethash 'P ht) #(10 3))
  (setf (gethash '[ ht) #(11 3))
  (setf (gethash '] ht) #(12 3))
  (setf (gethash 'backslash  ht) #(13 3))
  (setf (gethash 'forward-delete ht) #(15 3))
  (setf (gethash 'end ht) #(16 3))
  (setf (gethash 'page-up ht) #(17 3))
  (setf (gethash 'keypad-7 ht) #(19 3))
  (setf (gethash 'keypad-8 ht) #(20 3))
  (setf (gethash 'keypad-9 ht) #(21 3))
  (setf (gethash 'keypad-hyphen ht) #(22 3))
  ;; "A" row
  (setf (gethash 'caps-lock ht) #(0 4))
  (setf (gethash 'A ht) #(1 4))
  (setf (gethash 'S ht) #(2 4))
  (setf (gethash 'D ht) #(3 4))
  (setf (gethash 'F ht) #(4 4))
  (setf (gethash 'G ht) #(5 4))
  (setf (gethash 'H ht) #(6 4))
  (setf (gethash 'J ht) #(7 4))
  (setf (gethash 'K ht) #(8 4))
  (setf (gethash 'L ht) #(9 4))
  (setf (gethash 'semicolon ht) #(10 4))
  (setf (gethash 'quote ht) #(11 4))
  (setf (gethash 'return ht) #(12 4))
  (setf (gethash 'keypad-4 ht) #(19 4))
  (setf (gethash 'keypad-5 ht) #(20 4))
  (setf (gethash 'keypad-6 ht) #(21 4))
  (setf (gethash 'keypad-plus ht) #(22 4))
  ;; "Z" row
  (setf (gethash 'shift ht) #(0 5))
  (setf (gethash 'Z ht) #(1 5))
  (setf (gethash 'X ht) #(2 5))
  (setf (gethash 'C ht) #(3 5))
  (setf (gethash 'V ht) #(4 5))
  (setf (gethash 'B ht) #(5 5))
  (setf (gethash 'N ht) #(6 5))
  (setf (gethash 'M ht) #(7 5))
  (setf (gethash 'comma ht) #(8 5))
  (setf (gethash 'period ht) #(9 5))
  (setf (gethash 'dot  ht) #(9 5))
  (setf (gethash '/ ht) #(10 5))
  (setf (gethash 'right-shift ht) #(11 5))
  (setf (gethash 'up-arrow ht) #(16 5))
  (setf (gethash 'keypad-1 ht) #(19 5))
  (setf (gethash 'keypad-2 ht) #(20 5))
  (setf (gethash 'keypad-3 ht) #(21 5))
  (setf (gethash 'keypad-enter ht) #(22 5))
  ;; space bar row
  (setf (gethash 'left-control ht) #(0 6))
  (setf (gethash 'left-option ht) #(1 6))
  (setf (gethash 'left-command ht) #(2 6))
  (setf (gethash 'spc ht) #(3 6))
  (setf (gethash 'spc ht) #(4 6))
  (setf (gethash 'spc ht) #(5 6))
  (setf (gethash 'spc ht) #(6 6))
  (setf (gethash 'spc ht) #(7 6))
  (setf (gethash 'spc ht) #(8 6))
  (setf (gethash 'spc ht) #(9 6))
  (setf (gethash 'spc ht) #(10 6))
  (setf (gethash 'right-command ht) #(11 6))
  (setf (gethash 'right-option ht) #(12 6))
  (setf (gethash 'right-control ht) #(13 6))
  (setf (gethash 'left-arrow ht) #(15 6))
  (setf (gethash 'down-arrow ht) #(16 6))
  (setf (gethash 'right-arrow ht) #(17 6))
  (setf (gethash 'keypad-0 ht) #(19 6))
  (setf (gethash 'keypad-period ht) #(21 6))
  (setf (gethash 'enter ht) #(22 6))
  ht)


#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#