;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2004 Mike Byrne
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : motor-module.lisp
;;; Version     : 2.2
;;; 
;;; Description : Source code for the ACT-R/PM "motor module."
;;; 
;;; Bugs        : Issue of non-syncrhony between motor effective cursor pos
;;;             : and device cursor pos?  This true?
;;;
;;; ----- History -----
;;; 01.08.07 mdb
;;;             : Changed the QUEUE-COMMAND for a cursor-ply so that it
;;;             : doesn't call min-run-time for the MP.
;;; 02.03.05 mdb [b5]
;;;             : Fixed idiotic DISTANCE= and DIRECTION= bugs.  Duh.
;;; 02.03.08 mdb
;;;             : MOVE-CURSOR now checks to see if the hand is on the mouse.
;;; 2002.03.20 mdb
;;;             : Renamed accessor function TARGET-WIDTH to 
;;;             : DEFAULT-TARGET-WIDTH because that makes more sense (thanks,
;;;             : Frank R).
;;; 2002.06.27 mdb [b7]
;;;             : Moved CHECK-SPECS from here to PM-MODULE, which is were it
;;;             : should be.
;;; 2002.08.28 Dan [b8]
;;;             : Added Mike's updated move-cursor method that allows
;;;             : the action to be programmed while the hand is still 
;;;             : moving to the mouse from the previous action.
;;; 2003.09.10 mdb [2.1.3]
;;;             : * Bonnie John pointed out that hand homing was too slow.  
;;;             : This is because the target width used was too narrow (1 key).
;;;             : Changed this to 4.0, a more realistic width.
;;;             : * Also discovered that incremental mouse moves weren't 
;;;             : working due to a really old list/vector bug.  Grr. 
;;; 2003.10.23 mdb [2.1.3]
;;;             : Enforce separation between where the cursor effectively is
;;;             : for the Motor Module and where it acutally is according to
;;;             : the device.  [These may disagree if a movement is in progress
;;;             : and a subsequent move is being prepared.]
;;; 2004.03.06 mdb [2.2]
;;;             : Added switchable motor movement noise just to the cursor.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;




;;;; ---------------------------------------------------------------------- ;;;;
;;;; the Motor Module class itself
;;;; ---------------------------------------------------------------------- ;;;;


(defclass motor-module (pm-module)
  ((left-hand :accessor left-hand :initarg :left-hand
              :initform (make-instance 'hand))
   (right-hand :accessor right-hand :initarg :right-hand
               :initform (make-instance 'hand))
   (eff-cursor-loc :accessor eff-cursor-loc :initarg :eff-cursor-loc
                   :initform #(0 0))
   (feature-prep-time :accessor feat-prep-time  :initarg :feat-prep-time 
                      :initform 0.050)
   (movement-initiation-time :accessor init-time :initarg :init-time
                             :initform 0.050)
   (default-target-width :accessor default-target-width 
     :initarg :default-target-width :initform 1.0)
   (peck-fitts-coeff :accessor peck-fitts-coeff :initarg :peck-fitts-coeff
                     :initform 0.075) 
   (min-fitts-time :accessor min-fitts-time :initarg :min-fitts-time 
                   :initform 0.100)
   (incremental-mouse-p :accessor incremental-mouse-p 
                        :initarg :incremental-mouse-p :initform nil)
   (cursor-noise :accessor cursor-noise :initarg :cursor-noise
                 :initform nil)
   )
  (:default-initargs
    :version-string "2.2"
    :name :MOTOR))



(defmethod initialize-instance :after ((mtr-mod motor-module) &key)
  (home-hands mtr-mod)
  (setf (state-dmo mtr-mod)
        (make-dme 'motor-state 'module-state
                  '(module :motor modality free processor free preparation free
                    execution free)
                  :where :external)))


;;; RESET-MODULE      [Method]
;;; Date        : 97.03.31
;;; Description : Resetting a motor module includes clearing the motor
;;;             : feature memory and resetting the hands to the 'home row'
;;;             : position.

(defmethod reset-module :after ((mtr-mod motor-module))
  (setf (eff-cursor-loc mtr-mod) #(0 0))
  (home-hands mtr-mod))




;;;; ---------------------------------------------------------------------- ;;;;
;;;; the HAND class
;;;; ---------------------------------------------------------------------- ;;;;

(defclass hand ()
  ((location :accessor loc)
   (finger-offsets :accessor fingers :initarg :fingers)))


(defgeneric finger-loc (the-hand finger)
  (:documentation  "Return the absolute XY coordinate of <finger>"))

(defmethod finger-loc ((the-hand hand) finger)
  (let ((fngr-offset (second (assoc finger (fingers the-hand)))))
    (vector (+ (px (loc the-hand)) (px fngr-offset))
            (+ (py (loc the-hand)) (py fngr-offset)))))


(defgeneric move-finger (the-hand finger r theta)
  (:documentation  "Returns new XY loc of finger after a move of r, theta.  Also modifies finger offset."))

(defmethod move-finger ((the-hand hand) finger r theta)
  (unless (= theta 0.)
    (setf (second (assoc finger (fingers the-hand)))
          (polar-move-xy (second (assoc finger (fingers the-hand))) 
                         (vector r theta))))
  (finger-loc the-hand finger))


(defgeneric move-hand (the-hand r theta)
  (:documentation  "Moves the hand to a new location"))

(defmethod move-hand ((the-hand hand) r theta)
  (setf (loc the-hand) (polar-move-xy (loc the-hand) (vector r theta))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Support methods
;;;; ---------------------------------------------------------------------- ;;;;


#|  Not called by anything, oddly enough
(defmethod current-cursor-loc ()
  (get-mouse-coordinates (device (device-interface *mp*))))
|#

;;; MOVE-A-FINGER      [Method]
;;; Date        : 97.02.14
;;; Description : Moves a finger on the specified hand.

(defgeneric move-a-finger (mtr-mod hand finger r theta)
  (:documentation  "Moves a finger, returning the new XY location"))

(defmethod move-a-finger ((mtr-mod motor-module) hand finger r theta)
  (ecase hand
    (right (move-finger (right-hand mtr-mod) finger r theta))
    (left (move-finger (left-hand mtr-mod) finger r theta))))


;;; MOVE-A-HAND      [Method]
;;; Date        : 97.02.25
;;; Description : Moves a hand to the specified location

(defgeneric move-a-hand (mtr-mod hand r theta)
  (:documentation  "Moves a hand, returning the new XY location"))

(defmethod move-a-hand ((mtr-mod motor-module) hand r theta)
  (ecase hand
    (right (move-hand (right-hand mtr-mod) r theta))
    (left (move-hand (left-hand mtr-mod) r theta))))


;;; FINGER-LOC-M      [Method]
;;; Date        : 97.02.17
;;; Description : Module-level call to determine the location of a finger.

(defgeneric finger-loc-m (mtr-mod hand finger)
  (:documentation  "Return the XY location of the specified finger"))

(defmethod finger-loc-m ((mtr-mod motor-module) hand finger)
  (ecase hand
    (right (finger-loc (right-hand mtr-mod) finger))
    (left (finger-loc (left-hand mtr-mod) finger))))


;;; HOME-HANDS      [Method]
;;; Date        : 97.02.20
;;; Description : Sets the hand and finger locations to the keyboard home row.

(defgeneric home-hands (mtr-mod)
  (:documentation  "Sets the hand and finger locations to home row locations"))

(defmethod home-hands ((mtr-mod motor-module))
  (setf (loc (right-hand mtr-mod)) #(7 4))
  (setf (loc (left-hand mtr-mod)) #(4 4))
  (setf (fingers (right-hand mtr-mod))
        '((index #(0 0)) (middle #(1 0)) (ring #(2 0)) (pinkie #(3 0))
          (thumb #(-1 2))))
  (setf (fingers (left-hand mtr-mod))
        '((index #(0 0)) (middle #(-1 0)) (ring #(-2 0)) (pinkie #(-3 0))
          (thumb #(-1 2))))
  )




;;; LOC-TO-KEY      [Method]
;;; Date        : 97.02.26
;;; Description : Given a location, return the corresponding character.
;;;             : Accessed via the 'virtual keyboard' array.

(defgeneric loc->key (mtr-mod loc)
  (:documentation  "Given an location, return the corresponding key"))

(defmethod loc->key ((mtr-mod motor-module) (loc vector))
  (if (vpt= loc #(28 2))
    'mouse
    (aref (loc->key-arr mtr-mod) (px loc) (py loc))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Utility functions


(defun distance= (d1 d2)
  "Two distances are not equal if they are more than two degrees apart."
  (when (and d1 d2)
    (> 2.0 (abs (- d1 d2)))))


(defun direction= (d1 d2)
  "Two directions are not equal if they are more than pi/4 apart."
  (when (and d1 d2)
    (> (/ pi 4) (abs (- d1 d2)))))


(defun device->order (device)
  (ecase device
    (MOUSE 0)
    (JOYSTICK-1 1)
    (JOYSTICK-2 2)))


(defun check-specs (&rest specs)
  "If there is an invalid specification, return something, else NIL"
  (member nil specs))


(defgeneric xy-to-polar (from to)
  (:documentation  "Given a starting and ending location in xy, return polar difference"))

(defmethod xy-to-polar ((from vector) (to vector))
  (let ((distance (dist from to)))
    (vector distance
            (if (zerop distance)
              0.0
              (atan (- (py to) (py from)) 
                    (- (px to) (px from)))))))


(defmethod xy-to-polar ((from list) (to list))
  (xy-to-polar (coerce from 'vector) (coerce to 'vector)))


(defgeneric polar-move-xy (loc move)
  (:documentation  "Given an xy location and a polar displacement, return new xy"))

(defmethod polar-move-xy ((loc vector) (move vector))
  (round-xy
   (list (+ (px loc) (* (px move) (cos (py move))))
         (+ (py loc) (* (px move) (sin (py move)))))))

(defmethod polar-move-xy ((loc list) (move list))
  (polar-move-xy (coerce loc 'vector) (coerce move 'vector)))


;;; FITTS      [Function]
;;; Date        : 97.02.14, delta 99.06.18
;;; Description : Return Fitts law time for a given coefficient and distance,
;;;             : optionally supplying the target width (default is 1)

(defgeneric fitts (mtr-mod coef d &optional w)
  (:documentation  "Fitts law time for movement"))

(defmethod fitts ((mtr-mod motor-module) coef d &optional (w 1.0))
  (max (min-fitts-time mtr-mod)         ; 99.06.18
       (* coef (log (+ (/ d w) 0.5) 2))))



;;;; ---------------------------------------------------------------------- ;;;;
;;;; PUNCH class and methods

(defStyle punch () hand finger)


(defmethod compute-exec-time ((mtr-mod motor-module) (self punch))
  (+ (init-time mtr-mod) (key-closure-time (device-interface *mp*))))


(defmethod compute-finish-time ((mtr-mod motor-module) (self punch))
  (+ (init-time mtr-mod) (* 2 (burst-time mtr-mod))))


(defmethod feat-differences ((p1 punch) (p2 punch))
  (cond ((not (eq (hand p1) (hand p2))) 2)
        ((not (eq (finger p1) (finger p2))) 1)
        (t 0)))


(defmethod queue-output-events ((mtr-mod motor-module) (self punch))
  (queue-command 
   :where :DEVICE :command 'OUTPUT-KEY :time (exec-time self)
   :params
   (move-a-finger mtr-mod (hand self) (finger self) 0 0)))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; HAND-FINGER-R-THETA class and methods


(defStyle hfrt-movement () hand finger r theta)

(defmethod feat-differences ((m1 hfrt-movement) (m2 hfrt-movement))
  (let ((maxfeats (num-possible-feats m1)))
    (cond ((not (eq (hand m1) (hand m2))) (1- maxfeats))
          ((not (eq (finger m1) (finger m2))) (- maxfeats 2))
          (t
           (let ((nfeats 0))
             (unless (distance= (r m1) (r m2)) (incf nfeats))
             (unless (direction= (theta m1) (theta m2)) (incf nfeats))
             nfeats)))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; PECK class and methods

(defStyle peck hfrt-movement)

(defmethod compute-exec-time ((mtr-mod motor-module) (self peck))
  (+ (init-time mtr-mod)
     (max (burst-time mtr-mod)
          (rand-time 
           (fitts mtr-mod (peck-fitts-coeff mtr-mod) (r self))))))   ; 99.06.18


(defmethod queue-output-events ((mtr-mod motor-module) (self peck))
  (queue-command
   :where :DEVICE :command 'OUTPUT-KEY :time (exec-time self)
   :params
   (move-a-finger mtr-mod (hand self) (finger self) (r self) (theta self))))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; PECK-RECOIL class and methods

(defclass peck-recoil (hfrt-movement)
  ((move-time :accessor move-time))
  (:default-initargs
    :style-name :PECK))


(defmethod compute-exec-time ((mtr-mod motor-module) (self peck-recoil))
  (setf (move-time self)
        (max (burst-time mtr-mod)
             (fitts mtr-mod (peck-fitts-coeff mtr-mod) (r self))))   ; 99.06.18
  (+ (init-time mtr-mod)
     (max (burst-time mtr-mod) 
          (rand-time (move-time self)))))


(defmethod compute-finish-time ((mtr-mod motor-module) (self peck-recoil))
  (+ (exec-time self) (burst-time mtr-mod)
     (max (burst-time mtr-mod)
          (rand-time (move-time self)))))


(defmethod queue-output-events ((mtr-mod motor-module) (self peck-recoil))
  (queue-command
   :where :DEVICE :command 'OUTPUT-KEY :time (exec-time self)
   :params
   (polar-move-xy (finger-loc-m mtr-mod (hand self) (finger self))
                  (vector (r self) (theta self)))))


(defmethod peck-recoil ((mtr-mod motor-module) &key hand finger r theta)
  (unless (or (check-jam mtr-mod) (check-specs hand finger r theta))
    (when (symbolp theta)
      (setf theta (symbol-value theta)))
    (prepare-movement mtr-mod
                      (make-instance 'peck-recoil :hand hand :finger finger
                                     :r r :theta theta))))

;;;; ---------------------------------------------------------------------- ;;;;
;;;; PLY classes and methods

(defclass ply (hfrt-movement)
  ((fitts-coeff :accessor fitts-coeff :initarg :fitts-coeff)
   (target-width :accessor target-width :initarg :target-width
                 :initform (default-target-width (motor-m *mp*))))
  (:default-initargs
    :finger :dummy
    :style-name :PLY))


(defmethod compute-exec-time ((mtr-mod motor-module) (self ply))
  (+ (init-time mtr-mod)
     (rand-time (fitts mtr-mod (fitts-coeff self) (r self)
                       (target-width self)))))


(defmethod num-possible-feats :around ((self ply))
  (- (call-next-method) 1))


(defclass hand-ply (ply)
  ()
  (:default-initargs
    :fitts-coeff (when *mp* (mouse-fitts-coeff (device-interface *mp*)))))


(defmethod queue-output-events ((mtr-mod motor-module) (self hand-ply))
  (queue-command
   :where :MOTOR :command 'MOVE-A-HAND :time (exec-time self)
   :params (list (hand self) (r self) (theta self))))


(defclass cursor-ply (ply)
  ((target-coords :accessor target-coords :initarg :target-coords)
   (control-order :accessor control-order :initarg :control-order :initform 0))
  (:default-initargs
    :hand 'RIGHT
    :fitts-coeff (when *mp* (mouse-fitts-coeff (device-interface *mp*)))))


(defmethod compute-exec-time :before ((mtr-mod motor-module) 
                                         (self cursor-ply))
  (setf (fitts-coeff self) 
        (* (expt 2 (control-order self)) (fitts-coeff self))))


;;; QUEUE-OUTPUT-EVENTS      [Method]
;;; Date        : 99.04.02
;;; Description : New output queueing for cursor moves.  The move is broken
;;;             : up into pieces to provide the illusion of smooth movement.
;;;             : Queue up the pieces.  The last submovement doesn't move
;;;             : with a polar move, though, because small rounding errors
;;;             : accumulate and the target will be missed by a couple
;;;             : pixels.  

(defmethod queue-output-events ((mtr-mod motor-module) (self cursor-ply))
  (let* ((total-time (exec-time self))
         (nsteps (round total-time (min-run-time (cog-m *mp*))))
         (steplen nil))
    (when (incremental-mouse-p mtr-mod)
      (when (= nsteps 0) (setf nsteps 1))
      (setf steplen (/ (r self) nsteps))
      (when (> nsteps 1)
        (dotimes (idx (- nsteps 1))
          (queue-command 
           :command 'MOVE-CURSOR-POLAR :where :DEVICE
           :time (* (exec-time self) (/ idx nsteps))
           :params (list (vector 
                          (pm-angle-to-pixels steplen) (theta self)))))))
    (queue-command 
     :command 'MOVE-CURSOR-ABSOLUTE :where :DEVICE 
     :time (exec-time self) 
     :params (list (target-coords self)))
    (setf (eff-cursor-loc mtr-mod) (target-coords self))))


(defmethod point-hand ((mtr-mod motor-module) &key hand r theta twidth)
  (unless (or (check-jam mtr-mod) (check-specs hand r theta))
    (prepare-movement mtr-mod
                      (make-instance 'hand-ply :hand hand :r r :theta theta
                                     :target-width 
                                     (aif twidth it 
                                          (default-target-width mtr-mod))))))


;;; MOVE-CURSOR      [Method]
;;; Date        : 99.04.02
;;; Description : To get the movement time right, the effective width of the
;;;             : target has to be computed.  Gotta have the target object
;;;             : for that, which should be available through the supplied
;;;             : location or object.

(defmethod move-cursor ((mtr-mod motor-module) &key loc object
                          (device 'MOUSE))
  (unless (or (vpt= (loc (right-hand mtr-mod)) #(28 2))
              (and (eq (exec-s mtr-mod) 'BUSY) 
                   (eq (type-of (last-prep mtr-mod)) 'HAND-PLY)
                   (eq (hand (last-prep mtr-mod)) 'RIGHT)))
    (pm-warning "MOVE-CURSOR requested when hand not at mouse!")
    (return-from move-cursor nil))
  (unless (or (check-jam mtr-mod) (check-specs (or loc object) device))
    (let ((r-theta nil)
          (feat nil) (w nil))
      (setf feat
            (if object
              (obj-to-feat (vis-m *mp*) (psdme-to-dmo object))
              (loc-to-feat (vis-m *mp*) (psdme-to-dmo loc))))
      (setf r-theta (xy-to-polar (eff-cursor-loc mtr-mod) (xy-loc feat)))
      (unless (= 0 (vr r-theta))        ; r=0 is a no-op 
        (setf w (approach-width feat (vtheta r-theta)))
        (prepare-movement mtr-mod
                          (make-instance 'cursor-ply
                            :r (pm-pixels-to-angle (vr r-theta))
                            :theta (vtheta r-theta)
                            :target-width w
                            :target-coords (noisy-loc? mtr-mod (xy-loc feat) w)
                            :control-order (device->order device)))))))


(defgeneric noisy-loc? (mtr-mod xy-loc w)
  (:documentation "If the Motor Module is set up for it, make the output location noisy."))


;;; NOISY-LOC?      [Method]
;;; Description : Adds noise to the output location if noise is on.  Uses 
;;;             : logistic rather than normal [like other ACT-R].  7.8 is
;;;             : the width of the 96% interval of the unit logistic [after
;;;             : MacKenzie].

(defmethod noisy-loc? ((mm motor-module) (xy-loc vector) (w number))
  (if (not (cursor-noise mm))
    xy-loc
    (let ((pixw (pm-angle-to-pixels w)))
      (print pixw)
      (incf (px xy-loc) (round (noise (/ pixw 7.8))))
      (incf (py xy-loc) (round (noise (/ pixw 7.8))))
      xy-loc)))


;;;; ---------------------------------------------------------------------- ;;;;
;;;; Re-mapped methods


;;; CLICK-MOUSE      [Method]
;;; Date        : 97.02.13
;;; Description : Clicking the mouse is really just a punch with the right
;;;             : index finger.  

(defgeneric click-mouse (mtr-mod)
  (:documentation  "Execute a mouse click operation (a punch)"))

(defmethod click-mouse ((mtr-mod motor-module))
  (if (vpt= (loc (right-hand mtr-mod)) #(28 2))
    (punch mtr-mod :hand 'right :finger 'index)
    (pm-warning "CLICK-MOUSE requested when hand not at mouse!")))


;;; left in for backwards compatibility only
(defmethod move-mouse ((mtr-mod motor-module) location)
  (move-cursor mtr-mod :device 'mouse :loc location))


;;; PRESS-KEY      [Method]
;;; Date        : 97.02.18
;;; Description : Allows execution of key presses without computing movements
;;;             : to the keys and such.  Just does it via hash table lookup,
;;;             : all the 'intelligence' is in the hash table.

(defgeneric press-key (mtr-mod key)
  (:documentation  "High-level interface to press a key: Look up the command and execute it."))

(defmethod press-key ((mtr-mod motor-module) key)
  (when (stringp key)
    (setf key (read-from-string key)))
  (let ((command (key->cmd (device-interface *mp*) key)))
    (apply (first command) mtr-mod (rest command))))


(defgeneric point-hand-at-key (mtr-mod &key hand to-key)
  (:documentation  "Move the hand to a new location, specified by the key at which the index finger should point."))

(defmethod point-hand-at-key ((mtr-mod motor-module) &key hand to-key)
  (when (stringp to-key)
    (setf to-key (read-from-string to-key)))
  (let* ((cur-hand (ecase hand 
                       (right (right-hand mtr-mod))
                       (left (left-hand mtr-mod))))
         (new-polar (xy-to-polar (loc cur-hand) 
                                   (key-to-loc (keyboard (device-interface *mp*))
                                               to-key))))
    (point-hand mtr-mod :hand hand :r (vr new-polar) 
                :theta (vtheta new-polar))))


;;; HAND-TO-MOUSE      [Method]
;;; Date        : 97.02.25
;;; Description : Moves the right hand to the mouse.  Yes, I know the mouse
;;;             : location is hard-coded, but that's the way life is.  Just
;;;             : translates the movement into a POINT-HAND.

(defgeneric hand-to-mouse (mtr-mod)
  (:documentation  "Moves the right hand to the mouse"))

(defmethod hand-to-mouse ((mtr-mod motor-module))
  (unless (vpt= (loc (right-hand mtr-mod)) #(28 2))
    (let ((polar (xy-to-polar (loc (right-hand mtr-mod)) #(28 2))))
      (point-hand mtr-mod :hand 'right :r (vr polar) 
                  :theta (vtheta polar) :twidth 4.0))))


;;; HAND-TO-HOME      [Method]
;;; Date        : 97.02.25
;;; Description : Same thing as HAND-TO-MOUSE but in the other direction.

(defgeneric hand-to-home (mtr-mod)
  (:documentation  "Moves the right hand to the home row position from the mouse loc"))

(defmethod hand-to-home ((mtr-mod motor-module))
  (unless (equal (loc (right-hand mtr-mod)) #(7 4))
    (let ((polar (xy-to-polar (loc (right-hand mtr-mod)) #(7 4))))
      (point-hand mtr-mod :hand 'right :r (vr polar) 
                  :theta (vtheta polar) :twidth 4.0))))

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#