;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2004 Mike Byrne
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : rpm-parameters.lisp
;;; Version     : 2.1.3
;;; 
;;; Description : All the code that handles the global parameters for RPM.
;;;             : This is where default values are defined as well.
;;; 
;;; Bugs        : No known open bugs.
;;; 
;;; Todo        : Nothing pending.
;;; 
;;; ----- History -----
;;; 01.07.29 mdb
;;;             : 2.1/ACT5 work began
;;;             : [x] Remove obsolete :DISABLE-INTERRUPTS and 
;;;             :     :INTERRUPT-ON-OUTPUT parameters.
;;; 01.08.07 mdb
;;;             : Changed attn shift latency to 0.085 sec.
;;; 2002.03.20 mdb [b5]
;;;             : * Changed to accomodate new name for default target width
;;;             : accessor in motor module.  
;;;             : * Changed default of :VISUAL-MOVE-TOLERANCE from 0.0 to 0.5
;;;             : because it seems like there should be some by default.
;;; 2002.05.22 mdb [b6]
;;;             : Added :NEEDS-MOUSE parameter, removed :MOTOR-PROC-EVENTS.
;;; 2003.10.23 mdb [2.1.3]
;;;             : Added :REAL-TIME-SLACK-HOOK-FCT.
;;; 2004.03.12 mdb [2.2]
;;;             : Added :CURSOR-NOISE parameter.
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;; PM-PARAM      [Class]
;;; Date        : 97.03.07
;;; Description : Class for storing info about global PM parameters.

(defclass pm-param ()
  ((name :accessor name :initarg :name)
   (place :accessor place :initarg :place)
   (test :accessor restrict :initarg :test)
   (message :accessor msg :initarg :msg)
   (default :accessor default :initarg :default)))



;;; SET-PLACE      [Function]
;;; Date        : 97.03.07
;;; Description : Sets the value of PLACE, but evaluates PLACE first.
;;;             : Why the following macro doesn't work is still something
;;;             : of a mystery to me:
;;;             : (defmacro set-mac (place value)
;;;             :   `(setf ,place ,value))

(defun set-place (place value)
  "Sets <place> to <value>, but evaulates <place> first"
  (eval `(setf ,place ,value)))


;;; SET-PARAM      [Method]
;;; Date        : 97.03.07
;;; Description : Sets the value of a parameter, potentially.  First tests
;;;             : to see if the value matches the test supplied for that
;;;             : parameter.  If so, set it, if not, print a warning message.

(defgeneric set-param (pm-param value)
  (:documentation "Set a PM parameter to <value> assuming it passes the stored test."))

(defmethod set-param ((the-param pm-param) value)
  (if (funcall (restrict the-param) value)
    (set-place (place the-param) value)
    (pm-warning "Parameter ~S cannot take value ~A because it must be ~A."
                (name the-param) value (msg the-param))))


;;; RESTORE-DEFAULT      [Method]
;;; Date        : 98.01.12
;;; Description : Restore a parameter to its default value.

(defgeneric restore-default (pm-param)
  (:documentation "Restores a PM parameter to its default."))

(defmethod restore-default ((the-param pm-param))
  (set-place (place the-param) (default the-param)))


;;; PRINT-PARAM      [Method]
;;; Date        : 97.03.07
;;; Description : Prints the value of a parameter

(defgeneric print-param (pm-param)
  (:documentation "Prints the value of a PM parameter."))

(defmethod print-param ((the-param pm-param))
  (format t "~%Parameter ~S has value: ~A" (name the-param)
          (eval (place the-param))))



;;; BUILD-PARAM-LIST      [Function]
;;; Date        : 97.03.07
;;; Description : Builds the list of parameter objects.

(defun build-param-list ()
  "Build the list of global parameters"
  (list
   `(:audio-source-activation
     ,(make-instance 'pm-param :name :audio-source-activation :test #'posnum
                     :place '(source-act (audio-m *mp*)) :default 1.0
                     :msg "a non-negative number"))
   `(:auto-dequeue
     ,(make-instance 'pm-param :name :auto-dequeue :test #'tornil :default nil
                     :place '(auto-dequeue-p *mp*) :msg "T or NIL"))
   `(:conservative-update-visual
     ,(make-instance 'pm-param :name :conservative-update-visual :test #'tornil
                     :place '(conservative-p (vis-m *mp*)) :default nil
                     :msg "T or NIL"))
   `(:cursor-noise
     ,(make-instance 'pm-param :name :cursor-noise :test #'tornil
                     :place '(cursor-noise (motor-m *mp*)) :default nil
                     :msg "T or NIL"))
   `(:default-target-width
      ,(make-instance 'pm-param :name :default-target-width :test #'posnum
                      :place '(default-target-width (motor-m *mp*)) 
                      :msg "a non-negative number" :default 1.0))
   `(:device-updater-fct
     ,(make-instance 'pm-param :name :device-updater-fct :test #'functionp
                     :place '(device-hook (device-interface *mp*)) 
                     :msg "a function" :default nil))
   `(:digit-detect-delay
     ,(make-instance 'pm-param :name :digit-detect-delay :test #'posnum
                     :place '(digit-dtct-dly (audio-m *mp*)) 
                     :msg "a non-negative number" :default 0.3))
   `(:digit-duration
     ,(make-instance 'pm-param :name :digit-duration :test #'posnum
                     :place '(digit-duration (audio-m *mp*)) 
                     :msg "a non-negative number" :default 0.6))
   `(:digit-recode-delay
     ,(make-instance 'pm-param :name :digit-recode-delay :test #'posnum
                     :place '(digit-rec-dly (audio-m *mp*)) 
                     :msg "a non-negative number" :default 0.5))
   `(:incremental-mouse-moves
     ,(make-instance 'pm-param :name :incremental-mouse-moves :test #'tornil
                     :place '(incremental-mouse-p (motor-m *mp*)) :default NIL
                     :msg "T or NIL"))
   `(:min-fitts-time
     ,(make-instance 'pm-param :name :min-fitts-time :test #'posnum
                     :place '(min-fitts-time (motor-m *mp*)) 
                     :default 0.100  :msg "a non-negative number"))
   `(:motor-burst-time
     ,(make-instance 'pm-param :name :motor-burst-time :test #'posnum
                     :place '(burst-time (motor-m *mp*)) 
                     :msg "a non-negative number" :default 0.05))
   `(:motor-initiation-time
     ,(make-instance 'pm-param :name :motor-initiation-time :test #'posnum
                     :place '(init-time (motor-m *mp*)) 
                     :msg "a non-negative number" :default 0.05))
   `(:motor-feature-prep-time
     ,(make-instance 'pm-param :name :motor-feature-prep-time :test #'posnum
                     :place '(feat-prep-time (motor-m *mp*)) 
                     :msg "a non-negative number" :default 0.05))
   `(:mouse-fitts-coeff
     ,(make-instance 'pm-param :name :mouse-fitts-coeff :test #'posnum
                     :place '(mouse-fitts-coeff (device-interface *mp*)) 
                     :default 0.1  :msg "a non-negative number"))
   `(:needs-mouse
     ,(make-instance 'pm-param :name :needs-mouse :test #'tornil
                     :place '(needs-mouse-p (device-interface *mp*)) 
                     :default t  :msg "T or NIL."))
   `(:optimize-visual
     ,(make-instance 'pm-param :name :optimize-visual :test #'tornil :default T
                     :place '(optimize-p (vis-m *mp*)) :msg "T or NIL"))
   `(:output-speech
     ,(make-instance 'pm-param :name :output-speech :test #'tornil :default NIL
                     :place '(device-speech-p (device-interface *mp*)) 
                     :msg "T or NIL"))
   `(:peck-fitts-coeff
     ,(make-instance 'pm-param :name :peck-fitts-coeff :test #'posnum
                     :place '(peck-fitts-coeff (motor-m *mp*)) 
                     :default 0.075  :msg "a non-negative number"))
   `(:pixels-per-inch
     ,(make-instance 'pm-param :name :pixels-per-inch :test #'posnum
                     :place '(ppi (device-interface *mp*)) :default 72.0
                     :msg "a non-negative number"))
   `(:print-viewed
     ,(make-instance 'pm-param :name :print-viewed :test #'tornil :default T
                     :place '(print-view-p (vis-m *mp*)) :msg "T or NIL"))
   `(:process-cursor
     ,(make-instance 'pm-param :name :process-cursor :test #'tornil
                     :place '(with-cursor-p (device-interface *mp*))
                     :msg "T or NIL" :default NIL))
   `(:randomize-time
     ,(make-instance 'pm-param :name :randomize-time :test #'tornil :default NIL
                     :place '(randomize-time *mp*) :msg "T or NIL"))
   `(:real-time
     ,(make-instance 'pm-param :name :real-time :test #'tornil :default NIL
                     :place '(real-time-p *mp*) :msg "T or NIL"))
   `(:real-time-slack-hook-fct
     ,(make-instance 'pm-param :name :real-time-slack-hook-fct :test #'functionp
                     :place '(rt-slack-hook-fct *mp*) 
                     :msg "a function" :default nil))
   `(:show-focus
     ,(make-instance 'pm-param :name :show-focus :test #'tornil :default NIL
                     :place '(show-focus-p (device-interface *mp*)) 
                     :msg "T or NIL"))
   `(:sound-decay-time
     ,(make-instance 'pm-param :name :sound-decay-time :test #'posnum
                     :place '(decay-time (audio-m *mp*)) :default 3.0
                     :msg "a non-negative number"))
   `(:speech-hook-fct
     ,(make-instance 'pm-param :name :speech-hook-fct :test #'functionp
                     :place '(speech-hook (device-interface *mp*)) 
                     :msg "a function" :default nil))
   `(:syllable-rate
     ,(make-instance 'pm-param :name :syllable-rate :test #'posnum
                     :place '(s-rate (speech-m *mp*)) :default 0.15
                     :msg "a non-negative number"))
   `(:tone-detect-delay
     ,(make-instance 'pm-param :name :tone-detect-delay :test #'posnum
                     :place '(tone-dtct-dly (audio-m *mp*)) :default 0.05
                     :msg "a non-negative number"))
   `(:tone-recode-delay
     ,(make-instance 'pm-param :name :tone-recode-delay :test #'posnum
                     :place '(tone-rec-dly (audio-m *mp*)) :default 0.285
                     :msg "a non-negative number"))
   `(:trace-modules
     ,(make-instance 'pm-param :name :trace-modules :test #'tornil :default T
                     :place '(trace-modules *mp*) :msg "T or NIL"))
   `(:viewing-distance
     ,(make-instance 'pm-param :name :viewing-distance :test #'posnum
                     :place '(viewing-distance (device-interface *mp*)) 
                     :default 15.0  :msg "a non-negative number"))
   `(:visual-attention-latency
     ,(make-instance 'pm-param :name :visual-attention-latency :test #'posnum
                     :place '(move-attn-latency (vis-m *mp*)) :default 0.085
                     :msg "a non-negative number"))
   `(:visual-finst-span
     ,(make-instance 'pm-param :name :visual-finst-span :test #'posnum
                     :place '(finst-span (vis-m *mp*)) :default 3.0
                     :msg "a non-negative number"))
   `(:visual-movement-tolerance
     ,(make-instance 'pm-param :name :visual-movement-tolerance :test #'posnum
                     :place '(move-allowance (vis-m *mp*)) :default 0.5
                     :msg "a non-negative number"))
   `(:visual-num-finsts
     ,(make-instance 'pm-param :name :visual-num-finsts :test #'posnum
                     :place '(num-finst (vis-m *mp*)) :default 4
                     :msg "a non-negative number"))
   `(:visual-onset-span
     ,(make-instance 'pm-param :name :visual-onset-span :test #'posnum
                     :place '(new-span (vis-m *mp*)) :default 0.5
                     :msg "a non-negative number"))
   `(:visual-source-activation
     ,(make-instance 'pm-param :name :visual-source-activation :test #'posnum
                     :place '(source-act (vis-m *mp*)) :default 1.0
                     :msg "a non-negative number"))
   ))


#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#