;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Mike Byrne
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2004 Mike Byrne
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : rpm-toplevel.lisp
;;; Version     : 2.2
;;; 
;;; Description : Top-level functions for RPM, nothing fancy.
;;; 
;;; To do       :  
;;;
;;; ----- History -----
;;; 01.07.27 mdb
;;;             : Started 2.1/ACT5 conversion.
;;;             : [x] Parameterize buffer-stuffing
;;; 01.10.05 mdb [b3]
;;;             : Fixed bug in PM-SET-HAND-LOCATION.
;;; 01.11.05 mdb
;;;             : Restored :full-time in PM-RUN.
;;; 2002.04.16 mdb [b6]
;;;             : Made PM-PRINT-ICON more tabular.
;;; 02.05.06 Dan
;;;             : Updated the macros that didn't have a ` to have one, and 
;;;             : removed all the ,'s before *mp* in macros.
;;; 2002.10.18 mdb [b8]
;;;             : Added doc string to PM-SET-VISLOC-DEFAULT.
;;; 2003.02.03 mdb [2.1.1]
;;;             : Minor change to PM-RESET. 
;;; 2003.02.06 mdb
;;;             : Added module printing stuff to PM-VERSION. 
;;; 2003.10.23 mdb [2.1.3]
;;;             : Fixed cursor setting function for new handling.
;;; 2003.10.24 mdb
;;;             : Added PM-REGISTER-ARTICULATION-TIME.
;;; 2004.03.03 mdb [2.2]
;;;             : PM-ADD-SCREEN-OBJECT and PM-DELETE-SCREEN-OBJECT functions
;;;             : added.
;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defmacro pm-set-cursor-position (x y)
  "Sets the position of the cursor."
  `(pm-set-cursor-position-fct ,(vector x y)))


(defmacro pm-attend-location (x y)
  `(pm-attend-location-fct ,(vector x y)))


(defmacro pm-set-hand-location (hand &rest loc)
  "Sets the location of the given hand to LOC"
  `(pm-set-hand-location-fct ',hand ',loc))


(defmacro pm-set-params (&rest parameters)
  `(set-pm-params-mth *mp* ',parameters))


(defmacro pm-show-params (&rest parameters)
  `(show-pm-params-mth *mp* ',parameters))


(defmacro pm-prepare-motor (&rest lis)
  "Tells the Motor Module to prepare the supplied movement."
  `(pm-prepare-mvmt-mth (motor-m *mp*) ',lis))


(defun pm-add-screen-object (obj)
  (add-screen-object obj (vis-m *mp*)))


(defun pm-delete-screen-object (obj)
  (delete-screen-object obj (vis-m *mp*)))


(defun pm-reset (&key (initialize-dm t))
  "Resets the Master Process"
  (setf *pop-upon-failure* nil)
  ;(setf *external-sources* nil)
  (when initialize-dm
    (pm-initialize-dm))
  (reset-mp *mp*)
  )


(defun pm-add-chunks ()
  "Adds the chunks used by the PM layer"
  (add-pm-state-chunks)
)


(defun pm-add-types-and-chunks ()
  "Deprecated function from RPM 1.0"
  (pm-warning "PM-ADD-TYPES-AND-CHUNKS deprecated, use PM-RESET or PM-INITIALIZE-DM.")
  (pm-initialize-dm))


(defun pm-initialize-dm ()
  "Defines chunk types and adds chunks for the PM layer"
  (def-pm-chunk-types)
  (pm-add-chunks))


(defun pm-time ()
  "Return the current time, in seconds, of the Master Process"
  (mp-time *mp*))


(defun pm-proc-screen (&key clear with-cursor)
  "Processes the current display.  For backward compatibility only, use PM-PROC-DISPLAY."
  (when with-cursor
    (setf (with-cursor-p (device-interface *mp*)) t))
  (process-display (device-interface *mp*) (vis-m *mp*) clear))


(defun pm-proc-display (&key clear)
  "Processes the current display."
  (process-display (device-interface *mp*) (vis-m *mp*) clear))


(defun pm-install-window (window)
  "Installs <window> as the action window for the PM layer.  Included purely fo backward compatibility only. Use PM-INSTALL-DEVICE instead."
  (setf (device (device-interface *mp*)) window)
  (update-device (device-interface *mp*) (mp-time *mp*)))


(defun pm-install-device (device)
  "Installs <device> as the active device for the perceptual-motor layer."
  (setf (device (device-interface *mp*)) device))


(defun pm-pixels-to-angle (pixels)
  "Convert <pixels> to degress of visual angle."
  (pixels->angle-mth (device-interface *mp*) pixels))


(defun pm-angle-to-pixels (angle)
  "Convert visual <angle> in degress to pixels."
  (angle->pixels-mth (device-interface *mp*) angle))


(defun pm-set-cursor-position-fct (xyloc)
  (setf (eff-cursor-loc (motor-m *mp*)) xyloc
        (true-cursor-loc (device-interface *mp*)) xyloc)
  (synch-mouse (device-interface *mp*))
  (true-cursor-loc (device-interface *mp*)))


(defun pm-attend-location-fct (loc)
  "Tells the Vision Moddule to start with attention at a certain location."
  (setf (current-marker (vis-m *mp*)) (xy-to-dmo loc t))
  (setf (current-lof (vis-m *mp*)) loc))
       

(defun pm-remove-features (loc &optional scale)
  "Removes features at the specified visual location."
  (delete-features (vis-m *mp*) loc scale))


(defun pm-start-hand-at-mouse ()
  "Starts the right hand on the mouse instead of the default 'home row' location"
  (let ((the-hand (right-hand (motor-m *mp*))))
    (setf (loc the-hand) #(28 2))))


(defun pm-set-char-feature-set (setname)
  "Sets the feature set used to represent characters when optimizing is off. <setname> should be a keyword."
  (set-cfs-mth (vis-m *mp*) setname))


(defun pm-set-hand-location-fct (hand loc)
  "Function to set the location of the given hand to LOC"
  (setf loc (coerce loc 'vector))
  (ecase hand
    (right (setf (loc (right-hand (motor-m *mp*))) loc))
    (left (setf (loc (left-hand (motor-m *mp*))) loc))))
  

(defun pm-step ()
  "Enters the ACT-R/PM stepper"
  (step-rpm-mth *mp*))


(defun pm-run (duration &key (full-time nil))
  "Runs ACT-R/PM for the specifed duration (in seconds).
   If full-time is t then it gurantees that time is advanded by duration seconds"
  (when full-time
    (pm-delayed-event duration #'(lambda () 
                                   (pm-output nil "* Running stopped because time limit reached."))))
  
  (run-master-process *mp* duration))


(defun pm-print-icon ()
  "Print the Vision Manager's icon. For debugging."
  (format t "~%Loc        Att   Kind           Value             Color           ID")
  (format t "~%---------  ---   -------------  ----------------  --------------  -------------")
  (dolist (feat (visicon (vis-m *mp*)))
    (print-icon-feature feat)))


(defun pm-print-sched ()
  "Print the schedule queue. For debugging."
  (print-sched-queue *mp*))


(defun pm-print-input-q (module)
  "Print the input queue of the specified (by keyword) module.  For debugging."
  (print-input-queue (key->mgr *mp* module)))


(defun pm-print-module-state (module-name)
  "Print the state of the specified (by keyword) module."
  (print-mod-state-mth *mp* module-name))


(defun pm-version (&optional (print-modules t))
  "Returns the version string for RPM."
  (when print-modules
    (format *standard-output* "~%Overall system: ~A" (version-string *mp*))
    (map-modules *mp*
                 #'(lambda (mod)
                     (format *standard-output*
                             "~% Module ~S is version ~A" (my-name mod) 
                             (version-string mod)))))
  (version-string *mp*))


(defmacro pm-set-visloc-default (&rest params)
  "Set the default specification for visual location buffer stuffing."
  `(setf (default-spec (vis-m *mp*))
         (apply #'construct-findloc-spec ',params)))


(defun pm-register-articulation-time (string time)
  "Register the articulation time of a string."
  (register-art-time (speech-m *mp*) string time))


;;; left in for backward compatibilty only
(defun run-mp (duration)
  (run-master-process *mp* duration))

#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#