;;;  -*- mode: LISP; Package: CL-USER; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Authors     : Mike Byrne & Dan Bothell
;;; Address     : Rice University, MS-25
;;;             : Psychology Department
;;;             : Houston,TX 77251-1892
;;;             : byrne@acm.org
;;; 
;;; Copyright   : (c)1998-2003 Mike Byrne & Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LICENSE.txt
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : rpm-utils.lisp
;;; Version     : 2.2
;;; 
;;; Description : Utility macros and functions for ACT-R/PM, stuff that I 
;;;             : thought shouldn't be file-specific. 
;;; 
;;; Bugs        : None known.
;;; 
;;; Todo        : none
;;; 
;;; ----- History ----- 
;;; 97.01.15 Mike Byrne
;;;             : Genesis.  No notes kept for this file (dumb, sorry), until
;;;             : alpha 4.
;;; 97.06.20 Mike Byrne
;;;             : Added all the stuff for converting pixels to degrees and
;;;             : such.
;;; 97.06.25 Mike Byrne
;;;             : Added AWHEN.  Duh.
;;; 97.07.02 Mike Byrne
;;;             : Added MKSTR for the new string parsing in the Vision Manager.
;;; 98.01.12 Mike Byrne
;;;             : Added PM-OUTPUT and PM-WARNING macros.
;;; 98.01.21 Mike Byrne
;;;             : Added additional safety check in DME-TO-XY.
;;; 98.06.02 Mike Byrne
;;;             : Prepped for b3 release--many functions moved from here to
;;;             : where they should be (like "mcl-interface").
;;; 98.11.11 mdb
;;;             : b5 change, SYM->KEY utility function.
;;; 99.04.07 mdb
;;;             : Added OBJ-MATCH-* functions, commented out old screen crap.
;;; 99.07.06 mdb
;;;             : 1.0 changes.  Added a couple utils.
;;; 00.01.27 mdb
;;;             : Fixnum declaration in DEG->RAD changed to number for ACL.
;;; 99.12.21 Dan Bothell
;;;             : wrapped the while macro to avoid evaluation under ACL, 
;;;             : because a while function already exists.
;;; 00.06.08 mdb
;;;             : Started 2.0 work, added NEQ for ACL.
;;; 00.08.09 mdb
;;;             : Changed MATCH-SPEC-P to use vectors instead of lists.
;;; 00.08.10 mdb
;;;             : Started work to change internal representation of points 
;;;             : from lists to vectors.
;;; 01.03.26 mdb
;;;             : Added DEFGENERICs and doc strings.
;;; 01.05.31 Dan Bothell
;;;             : Changed random-item so that it doesn't call random on 0
;;;             : because that causes problems for LispWorks.  Interestingly
;;;             : (random 0) causes an error in ACL as well, but (random-item nil)
;;;             : in its old form didn't...weird.  [mdb note:  I changed this
;;;             : so it doesn't call LENGTH twice, which will be slow on a 
;;;             : long list]
;;; 01.06.15 Dan
;;;             : Removed the (declare (float x)) from ms-round
;;;             : because it was causing problems in CMUCL.
;;;             : It was being called with non-floats (in particular 
;;;             : things like 4/1000 by maybe-wait) and apparently
;;;             : the other Lisps just ignored it...  [mdb:  changed to NUBMER
;;;             : rather than FIXNUM].
;;; 01.06.15 Dan
;;;             : Put a feature check on pm-not-equal ((val fixnum))
;;;             : because lispworks and clisp don't like using fixnum as
;;;             : a class.
;;; 01.07.02 mdb
;;;             : Replaced PM-OUTPUT with Dan Bothell's NEW-PM-OUTPUT macro,
;;;             : to make things consistent.
;;; 
;;; 02.01.07 Dan 
;;;             : changed the fixnum declaration in rad->deg to number for ACL 
;;;             : (I can't believe I hadn't caught that when deg->rad was changed
;;;             : because without it that computation is really messed up which
;;;             : should have been noticed in approach width calculations...)
;;; 02.01.09 Dan
;;;             : added 2 more within methods to handle a mixture of variables
;;;             : (chunks) and numbers
;;; 02.01.15 Dan
;;;             : Changed the feature check on the defmacro of while
;;;             : because only ACL with the IDE has it, not every ACL.
;;; 2002.04.17 mdb [b6]
;;;             : Changed MS-ROUND to be less flaky, as per Dan's comment.
;;;
;;; 2002.08.14 Dan [b8]
;;;             : Changed the test on the while macro again because Franz has
;;;             : started putting it in their non-ide versions for versions 6+.
;;; 2002.10.01 Dan [2.1.1]
;;;             : Changed the pm-output macro so that it has *standard-output*
;;;             : instead of t in the formats because that seems to make a 
;;;             : difference when it gets called in a different process by the
;;;             : environment code (at least in ACL where there are multiple
;;;             : listener windows).
;;; 2003.02.06 mdb
;;;             : Added OBJS-NEAREST-SLOTVAL method.
;;; 2004.03.02 mdb [2.2]
;;;             : Made RANDOM-ITEM more generic.
;;; 2004.04.15 Dan
;;;             : Modified the more generic RANDOM-ITEM to work with ACL
;;;             : by specifying one for vector instead of simple-vector.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declaim (optimize (speed 3)))



;;;; ---------------------------------------------------------------------- ;;;;
;;;; Macros
;;;; ---------------------------------------------------------------------- ;;;;

;;; WHILE      [Macro]
;;; Description : While test is true, do body.  From _On Lisp_.

#-(or :allegro-ide (and :allegro-version>= (version>= 6)))
(defmacro while (test &body body)
  `(do ()
       ((not ,test))
     ,@body))


;;; AIF      [Macro]
;;; Date        : 97.02.09
;;; Description : From _On Lisp_, anaphoric if.  That is, can use variable
;;;             : "it" to refer to result of the test-form.

(defmacro aif (test-form then-form &optional else-form)
  `(let ((it ,test-form))
     (if it ,then-form ,else-form)))


(defmacro awhen (test-form &body body)
  `(aif ,test-form
     (progn ,@body)))



#-:mcl
(defmacro dovector ((varsym vec &optional ret) &body body)
  (let ((idx (gensym)))
    `(let (,varsym)
       (dotimes (,idx (length ,vec) ,ret)
         (setq ,varsym (aref ,vec ,idx))
         ,@body
         ))))


(defmacro px (vpt)
  "X coordinate of an XY vector."
  `(svref ,vpt 0))

(defmacro py (vpt)
  "Y coordinate of an XY vector."
  `(svref ,vpt 1))

(defmacro vr (vrt)
  "R component of an r-theta vector."
  `(svref ,vrt 0))

(defmacro vtheta (vrt)
  "Theta component of an r-theta vector."
  `(svref ,vrt 1))

;;;; ---------------------------------------------------------------------- ;;;;
;;;; Utility functions
;;;; ---------------------------------------------------------------------- ;;;;

#-:mcl
(defun neq (x y)
  "The NOT of EQ."
  (declare (inline neq))
  (not (eq x y)))

(defun vpt= (vpt1 vpt2)
  (and (= (px vpt1) (px vpt2))
       (= (py vpt1) (py vpt2))))

(defun sym->key (symbol)
  "Given a symbol, return the corresponding keyword."
  (read-from-string (mkstr ":" symbol)))


(defun key->sym (keyword)
  "Given a keyword, return the corresponding symbol."
  (read-from-string (mkstr keyword) 1))


;;; FLATTEN      [Function]
;;; Description : From Graham's _On Lisp_, takes a nested list and turns it
;;;             : into a flat one.  "Fast" version.

(defun flatten (lis)
  "Takes a nested list and makes in into a single-level list"
  (declare (list lis))
  (labels ((rec (lis acc)
             (cond ((null lis) acc)
                   ((atom lis) (cons lis acc))
                   (t (rec (car lis) (rec (cdr lis) acc))))))
    (rec lis nil)))


;;; MKLIST      [Function]
;;; Description : From Graham's _On Lisp_, make sure we have a list.

(defun mklist (obj)
  "If the object is not a list, return a list containing the object"
  (if (listp obj) obj (list obj)))


;;; MKSTR      [Function]
;;; Date        : 97.07.02
;;; Description : From Graham's _On Lisp_, makes sure we have a string.

(defun mkstr (&rest args)
  "Return a concatenated string representation of the arguments"
  (with-output-to-string (s)
    (dolist (a args) (princ a s))))


(defgeneric random-item (seq)
  (:documentation "Returns a random item from a sequence."))

(defmethod random-item ((seq list))
  (nth (random (length seq)) seq))

#+:mcl
(defmethod random-item ((seq simple-vector))
  (svref seq (random (length seq))))

#-:mcl
(defmethod random-item ((seq vector))
  (svref seq (random (length seq))))


(defmethod random-item ((seq sequence))
  (elt seq (random (length seq))))


;;; DIST      [Function]
;;; Description : Computes the distance between two locations (xy pairs, not
;;;             : chunks) using the  'real' hypoteneuse distance.

(defgeneric dist (loc1 loc2)
  (:documentation "Computes the distance in pixels between two XY locations"))

(defmethod dist ((loc1 vector) (loc2 vector))
  (sqrt (+ (expt (- (px loc1) (px loc2)) 2)
           (expt (- (py loc1) (py loc2)) 2))))

(defmethod dist ((loc1 list) (loc2 list))
  (dist (coerce loc1 'vector) (coerce loc2 'vector)))

;;; ROUND-XY      [Function]
;;; Date        : 97.02.13
;;; Description : Rounds an xy location to integer values

(defun round-xy (loc)
  (map 'vector #'round loc))

#| Why this doesn't work I'll never know...
(defun ms-round (x)
  "Rounds a time to the nearest millisecond"
  (declare (number x))
  (* 0.001 (round x 0.001)))
|#

(defun ms-round (x)
  "Rounds a time to the nearest millisecond"
  (declare (number x))
  (/ (round (* x 1000)) 1000.0))


(defun tornil (x)
  "Checks if a symbol is T or NIL"
  (or (eq x t) (eq x nil)))


(defun posnum (x)
  "Returns T only if <x> is a positive number"
  (and (numberp x) (>= x 0.)))


(defun rad->deg (r)
  "Converts radians into degrees."
  (declare (number r))
  (* r (/ 180 pi)))


(defun deg->rad (d)
  "Converts degrees into radians."
  (declare (number d))
  (* (/ pi 180) d))



(defmacro old-pm-output (message &rest args)
  `(when *verbose*
     (format t (concatenate 'string "~%<PM> " ,message) ,@args)))


(defmacro pm-output (time message &rest args)
  `(when *verbose*
     (if ,time
       (format *standard-output* "~%~VTTime ~6,3F: " *goal-depth* ,time)
       (format *standard-output*  "~%~VTTime ~6,3F: " *goal-depth* (mp-time *mp*)))
     (format *standard-output* ,message ,@args)))

(defmacro pm-warning (message &rest args)
  `(format *error-output* (concatenate 'string "~%#PM Warning# " ,message) 
           ,@args))



(defgeneric objs-match-slotval (lst slot-name value)
  (:documentation "Takes a list of CLOS objects and returns a list containing those items which have the slot <slot-name> equal to <value>."))

(defmethod objs-match-slotval ((ls list) (slot-name symbol) value)
  (when ls
    (let (accum)
      (dolist (obj ls (nreverse accum))
        (when (equal value (slot-value obj slot-name))
          (push obj accum))))))


(defmethod objs-match-slotval ((ls list) (slot-name symbol) 
                                  (value number))
  (when ls
    (let (accum)
      (dolist (obj ls (nreverse accum))
        (when (= value (slot-value obj slot-name))
          (push obj accum))))))


(defmethod objs-match-slotval ((ls list) (slot-name symbol) 
                                  (value symbol))
  (when ls
    (let (accum)
      (dolist (obj ls (nreverse accum))
        (when (eq value (slot-value obj slot-name))
          (push obj accum))))))


(defgeneric objs-min-slotval (lst slot-name)
  (:documentation "Given a list of CLOS objects and a slot name, return a list containing the object(s) with the lowest value for that slot."))

(defmethod objs-min-slotval ((ls list) (slot-name symbol))
  (when ls
    (let ((best (slot-value (first ls) slot-name))
          (current nil)
          (out-ls (list (first ls))))
      (dolist (obj (rest ls) (nreverse out-ls))
        (setf current (slot-value obj slot-name))
        (cond ((= current best) (push obj out-ls))
              ((< current best) 
               (setf best current)
               (setf out-ls (list obj))))))))


(defgeneric objs-max-slotval (lst slot-name)
  (:documentation "Given a list of CLOS objects and a slot name, return a list containing the object(s) with the highest value for that slot."))

(defmethod objs-max-slotval ((ls list) (slot-name symbol))
  (when ls
    (let ((best (slot-value (first ls) slot-name))
          (current nil)
          (out-ls (list (first ls))))
      (dolist (obj (rest ls) (nreverse out-ls))
        (setf current (slot-value obj slot-name))
        (cond ((= current best) (push obj out-ls))
              ((> current best) 
               (setf best current)
               (setf out-ls (list obj))))))))


(defgeneric objs-nearest-slotval (lst slot-name val)
  (:documentation "Given a list of CLOS objects and a slot name, return a list containing the objects with the slot value closest to the supplied value."))


(defmethod objs-nearest-slotval ((lst list) (slot-name symbol) 
                                     (val number))
  (let ((best (abs (- val (slot-value (first lst) slot-name))))
        (current nil)
        (out-lst  (list (first lst))))
    (dolist (obj (rest lst) (nreverse out-lst))
      (setf current (abs (- val (slot-value obj slot-name))))
      (cond ((= current best) (push obj out-lst))
            ((< current best)
             (setf best current)
             (setf out-lst (list obj)))))))



(defclass spec ()
  ((check-slots :accessor check-slots :initform nil :initarg :check-slots)))


(defmethod initialize-instance :after ((self spec) &key)
  (let (val)
    (dovector (slotname (check-slots self))        ; was dolist
      (setf val (slot-value self slotname))
      (when (and val (listp val) (symbol-function (first val)))
        (setf (slot-value self slotname) 
              (apply (first val) (rest val)))))))


(defgeneric match-spec-p (ts obj)
  (:documentation "Returns T if <obj> matches <spec>."))

(defmethod match-spec-p ((ts spec) (obj standard-object))
  (dovector (slotname (check-slots ts) t)
    (unless (slot-match-p ts obj slotname)
      (return-from match-spec-p nil))))


;;; SLOT-MATCH-P      [Function]
;;; Date        : 99.06.18
;;; Description : A slot matches the spec if the spec is :IGNORE, or if the
;;;             : the spec is a value and it EQUALs the value.  If the spec
;;;             : is a function, just return the result of the function
;;;             : call.

(defgeneric slot-match-p (spec object slotname)
  (:documentation "Determines if an object matches a spec on a particular slot."))

(defmethod slot-match-p ((ts spec) (obj standard-object)
                           (slotname symbol))
  (let ((condition (slot-value ts slotname))
        (value (slot-value obj slotname)))
    (cond ((eq condition :IGNORE) t)
          ((functionp condition)
           (funcall condition (slot-value obj slotname)))
          (t (equal condition value)))))


(defgeneric objs-match-spec (lst ts)
  (:documentation "Returns a list of objects from <lst> that match the spec <ts>."))

(defmethod objs-match-spec ((ls list) (ts spec))
  (remove-if #'(lambda (obj)
                 (not (match-spec-p ts obj))) ls))



(defgeneric pm-not-equal (val)
  (:documentation "Returns a closure that will evaluate if an argument is equal to <val>."))

;;; for symbols and integers (fixnums) we can use EQ.

(defmethod pm-not-equal ((val symbol))
  (lambda (s) (neq val s)))


#-(or :lispworks :clisp)
(defmethod pm-not-equal ((val fixnum))
  (lambda (s) (neq val s)))

;;; for floats, we have to round and then use EQ.

(defmethod pm-not-equal ((val float))
  (lambda (s) (not (eq (round s) (round val)))))

;;; for everything else, just use EQUAL.

(defmethod pm-not-equal (val)
  (setf val (check-gwn val))
  (lambda (s) (not (equal val s))))



(defgeneric within (min max)
  (:documentation "Returns a closure that will test whether the argument is betwen <min> and <max>, inclusive."))

(defmethod within ((min number) (max number))
  (lambda (val)
    (and (<= val max) (>= val min))))

(defmethod within ((min list) (max list))
  (within (check-fct min) (check-fct max)))

(defmethod within ((min number) (max list))
  (within min (check-fct max)))

(defmethod within ((min list) (max number))
  (within (check-fct min) max))


(defgeneric greater-than (criterion)
  (:documentation "Returns a closure that will return whether or not the argument is greater than <criterion>."))

(defmethod greater-than ((criterion number))
  (lambda (val)
    (> val criterion)))

(defmethod greater-than ((criterion list))
  (greater-than (check-fct criterion)))



(defgeneric less-than (criterion)
  (:documentation "Returns a closure that will return whether or not the argument is less than <criterion>."))

(defmethod less-than ((criterion number))
  (lambda (val)
    (< val criterion)))


(defmethod less-than ((criterion list))
  (less-than (check-fct criterion)))


(defgeneric string-to-lines (s)
  (:documentation "Given a string, break it into a list by #\Newline characters."))

(defmethod string-to-lines ((s string))
  (aif (position #\Newline s)
    (append (mklist (subseq s 0 it))
            (string-to-lines (subseq s (1+ it) (length s))))
    (list s)))


(defun check-gwn (x)
  "Checks to see if something is a GET-WME-NAME list.  If it is, it just evaluates it."
  (if (and (listp x) (eq (first x) 'GET-WME-NAME))
    (apply #'get-wme-name (list x))	;(eval x) mdb 01.07.28
    x))


(defgeneric check-fct (lst)
  (:documentation "If the input list represents a function call, return the value of that call.  Otherwise return the list."))

(defmethod check-fct ((ls list))
  (if (symbol-function (first ls))
    (eval ls)
    ls))


#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#